import os
import sys

# This demonstrates calling rotinv2dprepare in a limited memory environment so that you can experiment
# without stalling your session. On windows, the working set is limited and the program can page as much
# as usual. On linux, the address space has a hard limit and will terminate if the limit is exceeded.

# add the parent directory, so we find our iisignature build if it was built --inplace
sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
import iisignature

mem_limit_mb = 30
mem_limit_b = mem_limit_mb * 1024 * 1024

if os.name == "nt":
    import ctypes
    from ctypes import wintypes

    kernel32 = ctypes.WinDLL("kernel32", use_last_error=True)

    def errcheck_bool(result, func, args):
        if not result:
            raise ctypes.WinError(ctypes.get_last_error())
        return args

    kernel32.GetCurrentProcess.restype = wintypes.HANDLE
    kernel32.SetProcessWorkingSetSizeEx.errcheck = errcheck_bool
    kernel32.SetProcessWorkingSetSizeEx.argtypes = (
        wintypes.HANDLE,
        ctypes.c_size_t,
        ctypes.c_size_t,
        wintypes.DWORD,
    )
    hProcess = kernel32.GetCurrentProcess()
    kernel32.SetProcessWorkingSetSizeEx(hProcess, 204800, mem_limit_b, 6)
elif os.name == "posix":
    import resource

    resource.setrlimit(resource.RLIMIT_AS, (mem_limit_b, mem_limit_b))
else:
    raise RuntimeError("what system")

iisignature.rotinv2dprepare(14, "a")
print("ok")
# input()



# This depends on kivy being properly installed - install its dependencies first.
# It allows you to draw paths freehand with the mouse in the canvas, and
# see their log signatures printed on the console.
# based on https://kivy.org/docs/tutorials/firstwidget.html

from kivy.app import App
from kivy.uix.widget import Widget
from kivy.graphics import Color, Line
import numpy as np
import os
import sys

# add the parent directory, so we find our iisignature build if it was built --inplace
sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
import iisignature

np.set_printoptions(suppress=True)

m = 2
s = iisignature.prepare(2, m)


class MyPaintWidget(Widget):
    def on_touch_down(self, touch):
        with self.canvas:
            Color(1, 1, 0)
            touch.ud["line"] = Line(points=(touch.x, touch.y))

    def on_touch_move(self, touch):
        touch.ud["line"].points += [touch.x, touch.y]

    def on_touch_up(self, touch):
        path = np.reshape(touch.ud["line"].points, (-1, 2))
        print(iisignature.logsig(path, s))


class MyPaintApp(App):
    def build(self):
        return MyPaintWidget()


if __name__ == "__main__":
    MyPaintApp().run()



import os
import sys

# this so that we can import a test file from this directory
sys.path.append(os.path.abspath(os.path.dirname(__file__)))
import unittest
import test_sig


def my_module_suite():
    loader = unittest.TestLoader()
    suite = loader.loadTestsFromModule(test_sig)
    return suite



import iisignature
import tensorflow as tf
import numpy as np

# Provides Sig, LogSig, SigJoin and SigScale as tensorflow operations
# to match sig, logsig, sigjoin and sigscale from iisignature.
# Unlike for theano, there's no obvious reason to provide an
# op for siglength.

_zero = np.array(0.0, dtype="float32")


# SECTION 1: implementation of each gradient
def _sigGradImp(g, x, m):
    o = iisignature.sigbackprop(g, x, m)
    return o, _zero


class _logSigGradImp:
    def __init__(self, s, method):
        self.s = s
        self.method = method

    def __call__(self, g, x):
        return iisignature.logsigbackprop(g, x, self.s, self.method)


def _sigScaleGradImp(g, x, y, m):
    o = iisignature.sigscalebackprop(g, x, y, m)
    return o[0], o[1], _zero


def _sigJoinGradImp(g, x, y, m):
    o = iisignature.sigjoinbackprop(g, x, y, m)
    return o[0], o[1], _zero


def _sigJoinGradFixedImp(g, x, y, m, fixedlast):
    o = iisignature.sigjoinbackprop(g, x, y, m, fixedlast)
    return o[0], o[1], _zero, np.array(o[2], dtype="float32")


# SECTION 2: op for each gradient
def _sigGrad(op, grad):
    return tf.py_func(
        _sigGradImp, [grad] + list(op.inputs), [tf.float32] * 2, name="SigGrad"
    )


class _logSigGrad:
    def __init__(self, s, method):
        self.s = s
        self.method = method

    def __call__(self, op, grad):
        fn = _logSigGradImp(self.s, self.method)
        return tf.py_func(fn, [grad] + list(op.inputs), [tf.float32], name="LogSigGrad")


def _sigScaleGrad(op, grad):
    return tf.py_func(
        _sigScaleGradImp,
        [grad] + list(op.inputs),
        [tf.float32] * 3,
        name="SigScaleGrad",
    )


def _sigJoinGrad(op, grad):
    return tf.py_func(
        _sigJoinGradImp, [grad] + list(op.inputs), [tf.float32] * 3, name="SigJoinGrad"
    )


def _sigJoinGradFixed(op, grad):
    return tf.py_func(
        _sigJoinGradFixedImp,
        [grad] + list(op.inputs),
        [tf.float32] * 4,
        name="SigJoinGradFixed",
    )


# SECTION 3: implementation of each forward operation
def _sigImp(x, m):
    return iisignature.sig(x, m)


class _logSigImp:
    def __init__(self, s, method):
        self.s = s
        self.method = method

    def __call__(self, x):
        return iisignature.logsig(x, self.s, self.method)


def _sigScaleImp(x, y, m):
    return iisignature.sigscale(x, y, m)


def _sigJoinImp(x, y, m):
    return iisignature.sigjoin(x, y, m)


def _sigJoinFixedImp(x, y, m, fixedLast):
    return iisignature.sigjoin(x, y, m, fixedLast)


# SECTION 4: op for each forward operation
def Sig(x, m):
    rnd_name = "PyFuncGrad" + str(np.random.randint(0, 1e8))
    tf.RegisterGradient(rnd_name)(_sigGrad)
    g = tf.get_default_graph()
    with g.gradient_override_map({"PyFunc": rnd_name}):
        return tf.py_func(_sigImp, [x, m], tf.float32, name="Sig")


def LogSig(x, s, method=""):
    rnd_name = "PyFuncGrad" + str(np.random.randint(0, 1e8))
    tf.RegisterGradient(rnd_name)(_logSigGrad(s, method))
    g = tf.get_default_graph()
    with g.gradient_override_map({"PyFunc": rnd_name}):
        return tf.py_func(_logSigImp(s, method), [x], tf.float64, name="LogSig")


def SigScale(x, y, m):
    rnd_name = "PyFuncGrad" + str(np.random.randint(0, 1e8))
    tf.RegisterGradient(rnd_name)(_sigScaleGrad)
    g = tf.get_default_graph()
    with g.gradient_override_map({"PyFunc": rnd_name}):
        return tf.py_func(_sigScaleImp, [x, y, m], tf.float32, name="SigScale")


def SigJoin(x, y, m, fixedLast=None):
    rnd_name = "PyFuncGrad" + str(np.random.randint(0, 1e8))
    if fixedLast is None:
        tf.RegisterGradient(rnd_name)(_sigJoinGrad)
        g = tf.get_default_graph()
        with g.gradient_override_map({"PyFunc": rnd_name}):
            return tf.py_func(_sigJoinImp, [x, y, m], tf.float32, name="SigJoin")
    else:
        tf.RegisterGradient(rnd_name)(_sigJoinGradFixed)
        g = tf.get_default_graph()
        with g.gradient_override_map({"PyFunc": rnd_name}):
            return tf.py_func(
                _sigJoinFixedImp, [x, y, m, fixedLast], tf.float32, name="SigJoin"
            )


# https://stackoverflow.com/questions/37924071/tensorflow-writing-an-op-in-python
# https://gist.github.com/harpone/3453185b41d8d985356cbe5e57d67342



import sqlite3

# import grapher
# https://www.bastibl.net/publication-quality-plots/
import matplotlib as mpl

con = sqlite3.connect("mem_profile.sqlite")

if 0:
    # Check if there are duplicates
    q = "select count(*) as AA from A group by method,basis,d,m having AA>1"
    print(con.execute(q).fetchall())

    q = "select count(*)  from A"
    print(con.execute(q).fetchall())

    q = "select *  from A order by method,basis,d,m"
    all = con.execute(q).fetchall()
    i = 0
    for line in all:
        i += 1
        print(line)
    print(i)
    exit(0)


save = 1
if save:
    mpl.use("pdf")
import matplotlib.pyplot as plt  # noqa: E402

useLibertine = True
plt.rc("text", usetex=True)
if useLibertine:
    pre = r"""
\usepackage[T1]{fontenc}
\usepackage[tt=false,type1=true]{libertine}
%\setmonofont{inconsolata}
\usepackage[varqu]{zi4}
\usepackage[libertine]{newtxmath}
"""
    plt.rcParams["text.latex.preamble"] = (
        pre  #'\usepackage{libertine},\usepackage[libertine]{newtxmath},\usepackage{sfmath},\usepackage[T1]{fontenc}'
    )
else:
    plt.rc(
        "font",
        **{
            "family": "serif",
            "serif": ["Computer Modern"],
            "monospace": ["Computer Modern"],
        },
    )
plt.rc("xtick", labelsize=8)
plt.rc("ytick", labelsize=8)
plt.rc("legend", fontsize=8)
plt.rc("axes", labelsize=8)
mylinewidth = 0.6
plt.rc("axes", linewidth=mylinewidth)  # default 1

width = 3.487
height = width / 1.618
fig, ax = plt.subplots()
fig.subplots_adjust(left=0.15, bottom=0.16, right=0.99, top=0.97)

a = con.execute("select BYTES, D,M,METHOD,BASIS from A").fetchall()
b = [i for i in a]
bch_extra = 0


def get_basic(d, m, method, basis):
    found = False
    for i in b:
        usage = None
        if (d, m, method, basis) == (i[1], i[2], i[3], i[4]):
            if found:
                if usage != i[0]:
                    raise RuntimeError(
                        "two inconsistent copies of " + str((d, m, method))
                    )
            else:
                found = True
                usage = i[0]
    if not found:
        raise RuntimeError("no copies of " + str((d, m, method)))
    return usage


# bch_extra=get_basic(0,0,"C","L")
def get(d, m, method, basis):
    usage = get_basic(d, m, method, basis)
    if 0:
        if "C" == method:
            return usage - get_basic(0, 0, "C", "L")
        return usage
    return usage - get_basic(d, 1, method, basis)


# print (tabulate.tabulate(b,headers=["method","d","m", "reps","time"]))

Order = ["C", "S"]
d = 3
max_m = {2: 11, 3: 10, 4: 6, 5: 6}[d]
x = list(range(2, 1 + max_m))
y = [[get(d, m, method, "L") for m in x] for method in Order]
# series = [i for j in [[x,k,"+"] for k in y] for i in j]
# plt.plot(*series)

# d,x,o,8 and all the triangles look bigger , 8 and o look identical
# .,+,* look smaller
# These are the symbols and colours for the timings
# graph which uses the order C/O/S/esig
# so we just need elements 0 and 2
symbols = ["v", "o", "x", "d"]
colours = ["r", "b", "g", "k"]
symbols = (symbols[0], symbols[2])
colours = (colours[0], colours[2])
for i, nm, symbol, col in zip(y, Order, symbols, colours):
    # plt.plot(x,i,symbol,label=nm)
    nm1 = r"\verb|" + nm + "|"
    plt.scatter(x, i, c=col, marker=symbol, label=nm1, edgecolors="none")
# prop=mpl.font_manager.FontProperties("monospace")
legend = plt.legend(loc="upper left")
legend.get_frame().set_linewidth(mylinewidth)
plt.xlabel("level")
plt.yscale("log")
# plt.xlim(1.5,max_m+0.5)
plt.ylabel("usage(bytes) - logarithmic scale")
# grapher.pushplot(plt)
# plt.draw()
# dpi doesn't change the scale in the picture
filename = (
    "/home/jeremyr/Dropbox/phd/graphs/memsweep"
    + str(d)
    + "d"
    + ("Lib" if useLibertine else "")
)

if save:
    fig.set_size_inches(width, height)
    plt.savefig(filename + ".pdf")
else:
    plt.show()

# run this to check that the graph looks good in monochrome
if 0:
    # plt.savefig(filename+".png",dpi=300)
    from PIL import Image

    img = Image.open(filename + ".png").convert("LA")
    img.save(filename + "bw.png")



if __name__ != "__main__":
    import iisignature

import unittest
import numpy
import sys
import math
import time

numpy.set_printoptions(suppress=True, linewidth=150)
output_timings = False

CANCOMPILE = sys.platform != "darwin"


def prepare(d, m, methods):
    return iisignature.prepare(d, m, "".join(methodfilter(methods)))


def methodfilter(methods):
    """
    removes the unsupported methods from a list of method letters
    """
    if not CANCOMPILE:
        return [i for i in methods if i.upper() != "C"]
    return methods


HAVE_SCIPY = True
try:
    import scipy.linalg  # noqa: F401
except ImportError:
    HAVE_SCIPY = False


# text, index -> (either number or [res, res]), newIndex
def parseBracketedExpression(text, index):
    if text[index] == "[":
        left, m = parseBracketedExpression(text, index + 1)
        right, n = parseBracketedExpression(text, m + 1)
        return [left, right], n + 1
    else:
        n = 0
        while n < len(text) and text[index + n] in [
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
        ]:  # this should always just happen once if input is a bracketed expression of
            # letters
            n = n + 1
        return int(text[index : index + n]), index + n


# print (parseBracketedExpression("[23,[2,[[22,1],2]]]",0))


# bracketed expression, dim -> numpy array of its value, depth
def multiplyOut(expn, dim):
    if isinstance(expn, list):
        left, leftDepth = multiplyOut(expn[0], dim)
        right, rightDepth = multiplyOut(expn[1], dim)
        a = numpy.outer(left, right).flatten()
        b = numpy.outer(right, left).flatten()
        return a - b, leftDepth + rightDepth
    else:
        a = numpy.zeros(dim)
        a[expn - 1] = 1
        return a, 1


# string of bracketed expression, dim -> numpy array of its value, depth
# for example:
# valueOfBracket("[1,2]",2) is ([0,1,-1,0],2)
def valueOfBracket(text, dim):
    return multiplyOut(parseBracketedExpression(text, 0)[0], dim)


# inputs are values in the tensor algebra given as lists of levels (from 1 to
# level), assumed 0 in level 0.
# returns their concatenation product
def multiplyTensor(a, b):
    level = len(a)
    dim = len(a[0])
    sum = [numpy.zeros(dim**m) for m in range(1, level + 1)]
    for leftLevel in range(1, level):
        for rightLevel in range(1, 1 + level - leftLevel):
            sum[leftLevel + rightLevel - 1] += numpy.outer(
                a[leftLevel - 1], b[rightLevel - 1]
            ).flatten()
    return sum


# input is a value in the tensor algebra given as lists of levels (from 1 to
# level), assumed 0 in level 0.
# returns its exp - assumed 1 in level 0
# exp(x)-1 = x+x^2/2 +x^3/6 +x^4/24 + ...
def exponentiateTensor(a):
    out = [i.copy() for i in a]
    level = len(a)
    products = [out]
    for m in range(2, level + 1):
        t = multiplyTensor(a, products[-1])
        for j in t:
            j *= 1.0 / m
        products.append(t)
    return [numpy.sum([p[i] for p in products], 0) for i in range(level)]


# input is a value in the tensor algebra given as lists of levels (from 1 to
# level), assumed 1 in level 0.
# returns its log - assumed 0 in level 0
# log(1+x) = x -x^2/2 +x^3/3 -x^4/4 + ...
def logTensor(a):
    out = [i.copy() for i in a]
    level = len(a)
    products = [out]
    for m in range(2, level + 1):
        t = multiplyTensor(a, products[-1])
        products.append(t)
    neg = True
    for m in range(2, level + 1):
        for j in products[m - 1]:
            if neg:
                j *= -1.0 / m
            else:
                j *= 1.0 / m
        neg = not neg
    return [numpy.sum([p[i] for p in products], 0) for i in range(level)]


# given a tensor as a concatenated 1D array, return it as a list of levels
def splitConcatenatedTensor(a, dim, level):
    start = 0
    out = []
    for m in range(1, level + 1):
        levelLength = dim**m
        out.append(a[start : (start + levelLength)])
        start = start + levelLength
    assert start == a.shape[0]
    return out


# returns the signature of a straight line as a list of levels, and the number
# of multiplications used
def sigOfSegment(displacement, level):
    d = displacement.shape[0]
    sig = [displacement]
    mults = 0
    for m in range(2, level + 1):
        other = sig[-1]
        mults += 2 * other.shape[0] * d
        sig.append(numpy.outer(other, displacement).flatten() * (1.0 / m))
    return sig, mults


# inputs are values in the tensor algebra given as lists of levels (from 1 to
# level), assumed 1 in level 0.
# returns their concatenation product, and also the number of multiplications
# used
# c.f.  multiplyTensor
def chen(a, b):
    level = len(a)
    mults = 0
    sum = [a[m] + b[m] for m in range(level)]
    for leftLevel in range(1, level):
        for rightLevel in range(1, 1 + level - leftLevel):
            sum[leftLevel + rightLevel - 1] += numpy.outer(
                a[leftLevel - 1], b[rightLevel - 1]
            ).flatten()
            mults += a[leftLevel - 1].shape[0] * b[rightLevel - 1].shape[0]
    return sum, mults


def diff(a, b):
    return numpy.max(numpy.abs(a - b))


# Finite difference derivative
# estimate bump dot (the derivative of np.sum(f(X)) wrt X at X=x)
# i.e. the change in np.sum(f(X)) caused by bump
# with a finite difference approximation
def fdDeriv(f, x, bump, order, nosum=False):
    if order == 0:  # SIMPLE
        o = f(x + bump) - f(x)
    elif order == 2:  # 2nd order central
        o = 0.5 * (f(x + bump) - f(x - bump))
    elif order == 4:  # 4th order central
        o = (8 * f(x + bump) - 8 * f(x - bump) + f(x - 2 * bump) - f(x + 2 * bump)) / 12
    elif order == 6:  # 6th order central
        o = (
            45 * f(x + bump)
            - 45 * f(x - bump)
            + 9 * f(x - 2 * bump)
            - 9 * f(x + 2 * bump)
            + f(x + 3 * bump)
            - f(x - 3 * bump)
        ) / 60
    if nosum:
        return o
    return numpy.sum(o)


# fn takes an array to a scalar. This manually calculates its derivatives at init.
def allSensitivities(init, fn):
    z = numpy.zeros_like(init, dtype="float64")
    # y=fn(init)
    bump = 0.01
    o = numpy.zeros_like(init, dtype="float64")
    for i in range(len(z)):
        z[i] = bump
        # o[i]=(fn(init+z)-y)/bump
        # o[i]=(fn(init+z)-fn(init-z))/(2*bump)
        o[i] = fdDeriv(fn, init, z, 6) / bump
        z[i] = 0
    return o


class TestCase(unittest.TestCase):
    if output_timings:

        @classmethod
        def setUpClass(cls):
            cls.startTime = time.time()

        @classmethod
        def tearDownClass(cls):
            print(
                "\n%s.%s: %.3f"
                % (cls.__module__, cls.__name__, time.time() - cls.startTime)
            )


stack = numpy.stack


# numpy's lstsq has a different default in different versions
# I think this function agrees with newer versions (1.14+)
# and with scipy.
# This doesn't really affect the tests anyway.
def lstsq(a, b):
    rcond_to_use = max(a.shape) * numpy.finfo(float).eps
    return numpy.linalg.lstsq(a, b, rcond=rcond_to_use)


# This test checks that basis, logsig and sig are compatible with each other by
# calculating a signature both using sig
# and using logsig and checking they are equal
class A(TestCase):
    def consistency(self, coropa, dim, level):
        # numpy.random.seed(21)
        s = prepare(dim, level, "coshx" if coropa else "cosx")
        myinfo = {
            "level": level,
            "dimension": dim,
            "methods": "".join(methodfilter("COSAX" if level <= 2 else "COSX")),
            "basis": ("Standard Hall" if coropa else "Lyndon"),
            "logsigtosig_supported": False,
        }
        self.assertEqual(iisignature.info(s), myinfo)
        path = numpy.random.uniform(size=(10, dim))
        basis = iisignature.basis(s)
        logsig = iisignature.logsig(path, s)
        sig = iisignature.sig(path, level)

        # check lengths
        self.assertEqual(len(basis), iisignature.logsiglength(dim, level))
        self.assertEqual((len(basis),), logsig.shape)
        self.assertEqual(sig.shape, (iisignature.siglength(dim, level),))

        # calculate a signature from logsig
        expanded_logsig = [numpy.zeros(dim**m) for m in range(1, level + 1)]
        for coeff, expression in zip(logsig, basis):
            values, depth = valueOfBracket(expression, dim)
            expanded_logsig[depth - 1] += values * coeff
        calculated_sig = numpy.concatenate(exponentiateTensor(expanded_logsig))
        self.assertLess(diff(sig, calculated_sig), 0.00001)

        # calculate a log signature from sig
        fullLogSig = numpy.concatenate(
            logTensor(splitConcatenatedTensor(sig, dim, level))
        )
        fullLogSigLib = iisignature.logsig(path, s, "x")
        diff1 = numpy.max(numpy.abs(fullLogSigLib - fullLogSig))
        # print
        # (numpy.vstack([fullLogSig,fullLogSigLib,numpy.abs(fullLogSigLib-fullLogSig)]).transpose())
        self.assertLess(diff1, 0.00001)

        basisMatrix = []
        zeros = [numpy.zeros(dim**m) for m in range(1, level + 1)]
        for expression in basis:
            values, depth = valueOfBracket(expression, dim)
            temp = zeros[depth - 1]
            zeros[depth - 1] = values
            basisMatrix.append(numpy.concatenate(zeros))
            zeros[depth - 1] = temp
        calculatedLogSig = lstsq(numpy.transpose(basisMatrix), fullLogSig)[0]
        diff2 = numpy.max(numpy.abs(logsig - calculatedLogSig))
        self.assertLess(diff2, 0.00001)

        # check consistency of methods
        slowLogSig = iisignature.logsig(path, s, "o")
        diffs = numpy.max(numpy.abs(slowLogSig - calculatedLogSig))
        self.assertLess(diffs, 0.00001)

        sigLogSig = iisignature.logsig(path, s, "s")
        diffs = numpy.max(numpy.abs(sigLogSig - calculatedLogSig))
        self.assertLess(diffs, 0.00001)

        if level < 3:
            areaLogSig = iisignature.logsig(path, s, "a")
            diffs = numpy.max(numpy.abs(areaLogSig - calculatedLogSig))
            self.assertLess(diffs, 0.00001)

    def testConsistency(self):
        self.consistency(False, 3, 6)
        self.consistency(False, 3, 2)

    def testCoropa(self):
        self.consistency(True, 5, 2)

    # test sigjoin is compatible with sig and also its deriv
    def testjoining(self):
        numberToDo = 1
        dim = 2
        level = 2
        siglength = iisignature.siglength(dim, level)
        for fixedPoint, inputDim, fixed in [
            (float("nan"), dim, False),
            (0.1, dim - 1, True),
        ]:
            pathLength = 10

            def makePath():
                p = numpy.random.uniform(size=(pathLength, dim))
                if fixed:
                    p[:, -1] = fixedPoint * numpy.arange(pathLength)
                return p

            paths = [makePath() for i in range(numberToDo)]
            sig = numpy.vstack([iisignature.sig(path, level) for path in paths])

            joinee = numpy.zeros((numberToDo, siglength))
            for i in range(1, pathLength):
                displacements = [
                    path[i : (i + 1), :] - path[(i - 1) : i, :] for path in paths
                ]
                displacement = numpy.vstack(displacements)
                if fixed:
                    displacement = displacement[:, :-1]
                joinee = iisignature.sigjoin(joinee, displacement, level, fixedPoint)
            self.assertLess(
                diff(sig, joinee),
                0.0001,
                "fullSig matches sig" + (" with fixed Dim" if fixed else ""),
            )

            extra = numpy.random.uniform(size=(numberToDo, inputDim))
            bumpedExtra = 1.001 * extra
            bumpedJoinee = 1.001 * joinee
            base = numpy.sum(iisignature.sigjoin(joinee, extra, level, fixedPoint))
            bump1 = numpy.sum(
                iisignature.sigjoin(bumpedJoinee, extra, level, fixedPoint)
            )
            bump2 = numpy.sum(
                iisignature.sigjoin(joinee, bumpedExtra, level, fixedPoint)
            )
            derivsOfSum = numpy.ones((numberToDo, siglength))
            calculated = iisignature.sigjoinbackprop(
                derivsOfSum, joinee, extra, level, fixedPoint
            )
            self.assertEqual(len(calculated), 3 if fixed else 2)
            diff1 = (bump1 - base) - numpy.sum(calculated[0] * (bumpedJoinee - joinee))
            diff2 = (bump2 - base) - numpy.sum(calculated[1] * (bumpedExtra - extra))
            # print ("\n",bump1,bump2,base,diff1,diff2)
            self.assertLess(
                numpy.abs(diff1),
                0.000001,
                "diff1 as expected " + (" with fixed Dim" if fixed else ""),
            )
            self.assertLess(
                numpy.abs(diff2),
                0.00001,
                "diff2 as expected " + (" with fixed Dim" if fixed else ""),
            )
            if fixed:
                bumpedFixedPoint = fixedPoint * 1.01
                bump3 = numpy.sum(
                    iisignature.sigjoin(joinee, extra, level, bumpedFixedPoint)
                )
                diff3 = (
                    bump3
                    - base
                    - numpy.sum(calculated[2] * (bumpedFixedPoint - fixedPoint))
                )
                # print("\n",bump3,base, fixedPoint, bumpedFixedPoint, calculated[2])
                self.assertLess(numpy.abs(diff3), 0.00001, "diff3")

    # test sigcombine is compatible with sig and also its deriv
    def testcombining(self):
        dim = 2
        level = 2
        siglength = iisignature.siglength(dim, level)
        pathLength = 20
        halfPathLength = 10
        numberToDo = 4
        path = numpy.random.uniform(size=(numberToDo, pathLength, dim))
        sig = iisignature.sig(path, level)
        sig1 = iisignature.sig(path[:, :halfPathLength], level)
        sig2 = iisignature.sig(path[:, (halfPathLength - 1) :], level)
        combined = iisignature.sigcombine(sig1, sig2, dim, level)
        self.assertLess(diff(sig, combined), 0.0001)

        bumpedsig1 = 1.001 * sig1
        bumpedsig2 = 1.001 * sig2
        base = numpy.sum(iisignature.sigcombine(sig1, sig2, dim, level))
        bump1 = numpy.sum(iisignature.sigcombine(bumpedsig1, sig2, dim, level))
        bump2 = numpy.sum(iisignature.sigcombine(sig1, bumpedsig2, dim, level))
        derivsOfSum = numpy.ones((numberToDo, siglength))
        calculated = iisignature.sigcombinebackprop(derivsOfSum, sig1, sig2, dim, level)
        self.assertEqual(len(calculated), 2)
        diff1 = (bump1 - base) - numpy.sum(calculated[0] * (bumpedsig1 - sig1))
        diff2 = (bump2 - base) - numpy.sum(calculated[1] * (bumpedsig2 - sig2))
        # print ("\n",bump1,bump2,base,diff1,diff2)
        self.assertLess(numpy.abs(diff1), 0.000001)
        self.assertLess(numpy.abs(diff2), 0.00001)


class LogSig2Sig(TestCase):
    def doL2S(self, coropa):
        numpy.random.seed(212)
        d = 3
        m = 6
        path = numpy.random.uniform(size=(12, d))
        sig = iisignature.sig(path, m)
        s = iisignature.prepare(d, m, "S2H" if coropa else "S2")
        self.assertTrue(iisignature.info(s)["logsigtosig_supported"])
        logsig = iisignature.logsig(path, s)
        sig_ = iisignature.logsigtosig(logsig, s)
        self.assertEqual(sig.shape, sig_.shape)
        self.assertTrue(numpy.allclose(sig, sig_))

        # Like the other iisig functions, we check that derivatives
        # of logsigtosig allow sum(logsigtosig) to be backproped correctly
        # This is a boring thing to calculate, because after the first level
        # each of the log signature elements is a lie bracket and so
        # contributes a net total of 0 to the signature
        derivsOfSum = numpy.ones((sig.shape[0],), dtype="float64")
        bumpedLogSig = 1.01 * logsig
        calculated = iisignature.logsigtosigbackprop(derivsOfSum, logsig, s)
        # wantedbackprop = allSensitivities(logsig, lambda l: iisignature.logsigtosig(l,s).sum())
        manualChange = fdDeriv(
            lambda x: iisignature.logsigtosig(x, s), logsig, bumpedLogSig - logsig, 6
        )
        calculatedChange = numpy.sum((bumpedLogSig - logsig) * calculated)
        self.assertLess(numpy.abs(manualChange - calculatedChange), 0.00001)
        # beyond the first level, all zero
        if m > 1:
            self.assertLess(numpy.max(numpy.abs(calculated[d:])), 0.00001)
        self.assertEqual(calculated.shape, logsig.shape)

        # Now for a better test, we backprop sum(random*logsigtosig)
        # specifically calculate the change in it caused by bump two ways
        random = numpy.random.uniform(
            size=sig.shape[0],
        )
        derivsOfSum = random
        calculated = iisignature.logsigtosigbackprop(derivsOfSum, logsig, s)
        manualChange = fdDeriv(
            lambda x: iisignature.logsigtosig(x, s) * random,
            logsig,
            bumpedLogSig - logsig,
            4,
        )
        calculatedChange = numpy.sum((bumpedLogSig - logsig) * calculated)
        self.assertLess(numpy.abs(manualChange - calculatedChange), 0.00001)
        self.assertEqual(calculated.shape, logsig.shape)

    def testLyndon(self):
        self.doL2S(False)

    def testCoropa(self):
        self.doL2S(True)

    def testIssue16(self):
        # github-reported crash, issue #16
        iisignature.prepare(3, 3, "2")


class Deriv(TestCase):
    def testSig(self):
        # test that sigjacobian and sigbackprop compatible with sig
        numpy.random.seed(291)
        d = 3
        m = 5
        pathLength = 10
        path = numpy.random.uniform(size=(pathLength, d))
        path = numpy.cumsum(
            2 * (path - 0.5), 0
        )  # makes it more random-walk-ish, less like a scribble
        increment = 0.01 * numpy.random.uniform(size=(pathLength, d))
        base_sig = iisignature.sig(path, m)

        target = fdDeriv(
            lambda x: iisignature.sig(x, m), path, increment, 2, nosum=True
        )

        gradient = iisignature.sigjacobian(path, m)
        calculated = numpy.tensordot(increment, gradient)

        diffs = numpy.max(numpy.abs(calculated - target))
        niceOnes = numpy.abs(calculated) > 1.0e-4
        niceOnes2 = numpy.abs(calculated) < numpy.abs(base_sig)
        diffs1 = numpy.max(
            numpy.abs((calculated[niceOnes2] - target[niceOnes2]) / base_sig[niceOnes2])
        )
        diffs2 = numpy.max(numpy.abs(calculated[1 - niceOnes2] - target[1 - niceOnes2]))
        ratioDiffs = numpy.max(numpy.abs(calculated[niceOnes] / target[niceOnes] - 1))

        # numpy.set_printoptions(suppress=True,linewidth=os.popen('stty size',
        #'r').read().split()[1] #LINUX
        # numpy.set_printoptions(suppress=True,linewidth=150)
        # print ("")
        # print (path)
        # print
        # (numpy.vstack([range(len(base_sig)),base_sig,calculated,target,(calculated-target)/base_sig,calculated/target-1]).transpose())
        # print (diffs, ratioDiffs, diffs1, diffs2)
        # print(numpy.argmax(numpy.abs(calculated[niceOnes]/target[niceOnes]-1)),numpy.argmax(numpy.abs((calculated-target)/base_sig)))

        self.assertLess(diffs, 0.00001)
        self.assertLess(ratioDiffs, 0.01)
        self.assertLess(diffs1, 0.001)
        self.assertLess(diffs2, 0.00001)

        # compatibility between sigbackprop and sigjacobian is strong
        dFdSig = numpy.random.uniform(size=(iisignature.siglength(d, m),))
        backProp = iisignature.sigbackprop(dFdSig, path, m)
        manualCalcBackProp = numpy.dot(gradient, dFdSig)
        backDiffs = numpy.max(numpy.abs(backProp - manualCalcBackProp))
        if 0:  # to investigate the compile logic problem I used this and
            # (d,m,pathLength)=(1,2,2)
            print("")
            print(dFdSig)
            print(path)
            print(backProp)
            print(manualCalcBackProp)
        self.assertLess(backDiffs, 0.000001)

    def logSig(self, type, m=7):
        numpy.random.seed(291)
        d = 2
        pathLength = 10
        s = iisignature.prepare(d, m, type)
        path = numpy.random.uniform(size=(pathLength, d))
        path = numpy.cumsum(
            2 * (path - 0.5), 0
        )  # makes it more random-walk-ish, less like a scribble
        increment = 0.01 * path
        increment = 0.1 * numpy.random.uniform(size=(pathLength, d))

        manualChange = fdDeriv(
            lambda x: iisignature.logsig(x, s, type), path, increment, 4
        )

        dFdlogSig = numpy.ones(
            iisignature.siglength(d, m)
            if "X" == type
            else iisignature.logsiglength(d, m)
        )
        calculatedChange = numpy.sum(
            increment * iisignature.logsigbackprop(dFdlogSig, path, s, type)
        )
        # print(manualChange, calculatedChange)
        self.assertLess(numpy.abs(manualChange - calculatedChange), 0.0001)

    def testLogSig_expanded(self):
        self.logSig("X")

    def testLogSig_lyndon(self):
        self.logSig("S")

    def testLogSig_hall(self):
        self.logSig("H")

    def testLogSig_area(self):
        self.logSig("A", 2)

    def test_logsigbackwards_can_augment_s(self):
        numpy.random.seed(291)
        d = 2
        m = 7
        pathLength = 3
        path = numpy.random.uniform(size=(pathLength, d))
        dFdlogSig = numpy.ones(iisignature.logsiglength(d, m))
        for types in (("x", "o", "s"), ("xh", "oh", "sh")):
            ss = [iisignature.prepare(d, m, t) for t in types]
            backs = [iisignature.logsigbackprop(dFdlogSig, path, s) for s in ss]
            self.assertTrue(numpy.allclose(backs[0], backs[2]), types[0])
            self.assertTrue(numpy.allclose(backs[1], backs[2]), types[1])
            fwds = [iisignature.logsig(path, s, "s") for s in ss]
            self.assertTrue(numpy.allclose(fwds[0], fwds[2]), types[0])
            self.assertTrue(numpy.allclose(fwds[1], fwds[2]), types[1])


class LogSigJoin(TestCase):
    def testjoining(self):
        for level, method in [(3, "C"), (3, "O"), (2, "A")]:
            if method == "C" and not CANCOMPILE:
                continue
            numberToDo = 4
            dim = 5
            s = iisignature.prepare(dim, level, method)
            siglength = iisignature.logsiglength(dim, level)
            pathLength = 10

            def makePath():
                p = numpy.random.uniform(size=(pathLength, dim))
                return p

            paths = [makePath() for i in range(numberToDo)]
            sig = numpy.vstack([iisignature.logsig(path, s) for path in paths])

            joinee = numpy.zeros((numberToDo, siglength))
            for i in range(1, pathLength):
                displacements = [
                    path[i : (i + 1), :] - path[(i - 1) : i, :] for path in paths
                ]
                displacement = numpy.vstack(displacements)
                joinee = iisignature.logsigjoin(joinee, displacement, s)
            self.assertLess(diff(sig, joinee), 0.0001)

    def test_d(self):
        numberToDo = 20
        dim = 2
        pathLength = 10

        def makePath():
            p = numpy.random.uniform(size=(pathLength, dim))
            return p

        paths = [makePath() for i in range(numberToDo)]
        for level, method in [(2, "A"), (5, "O")]:
            siglength = iisignature.logsiglength(dim, level)
            s = iisignature.prepare(dim, level, method)

            sig = numpy.vstack([iisignature.logsig(path, s) for path in paths])
            joinee = sig
            extra = numpy.random.uniform(size=(numberToDo, dim))
            bumpedExtra = 1.001 * extra
            bumpedJoinee = 1.001 * joinee
            base = numpy.sum(iisignature.logsigjoin(joinee, extra, s))
            bump1 = numpy.sum(iisignature.logsigjoin(bumpedJoinee, extra, s))
            bump2 = numpy.sum(iisignature.logsigjoin(joinee, bumpedExtra, s))
            derivsOfSum = numpy.ones((numberToDo, siglength))
            calculated = iisignature.logsigjoinbackprop(derivsOfSum, joinee, extra, s)

            self.assertEqual(len(calculated), 2)
            diff1 = (bump1 - base) - numpy.sum(calculated[0] * (bumpedJoinee - joinee))
            diff2 = (bump2 - base) - numpy.sum(calculated[1] * (bumpedExtra - extra))
            self.assertLess(numpy.abs(diff1), 0.000001)
            self.assertLess(numpy.abs(diff2), 0.00001)


class Counts(TestCase):
    # check sigmultcount, and also that sig matches a manual signature
    # calculation
    def testa(self):
        numpy.random.seed(2141)
        d = 5
        m = 5
        pathLength = 4
        displacement = numpy.random.uniform(size=(d))
        sig, mults = sigOfSegment(displacement, m)
        path = [numpy.zeros(d), displacement]
        for x in range(pathLength):
            displacement1 = numpy.random.uniform(size=(d))
            path.append(path[-1] + displacement1)
            sig1, mults1 = sigOfSegment(displacement1, m)
            sig, mults2 = chen(sig, sig1)
            mults += mults1 + mults2
        # print (path)
        path = numpy.vstack(path)
        isig = iisignature.sig(path, m, 1)
        for i in range(m):
            self.assertLess(diff(isig[i], sig[i]), 0.00001)
        self.assertEqual(mults, iisignature.sigmultcount(path, m))


class SimpleCases(TestCase):
    def testFewPoints(self):
        # check sanity of paths with less than 3 points
        path1 = [[4.3, 0.8]]
        path2 = numpy.array([[1, 2], [2, 4]])
        m = 4
        d = 2
        s = prepare(d, m, "cosx")
        s_a = prepare(d, 2, "cosx")
        length = iisignature.siglength(d, m)
        loglength = iisignature.logsiglength(d, m)
        loglength_a = iisignature.logsiglength(d, 2)
        blankLogSig = numpy.zeros(loglength)
        blankLogSig_a = numpy.zeros(loglength_a)
        blankSig = numpy.zeros(length)
        self.assertLess(diff(iisignature.sig(path1, m), blankSig), 0.000000001)
        self.assertTrue(
            numpy.array_equal(iisignature.sig(path1, m, 2), numpy.zeros([0, length]))
        )
        if CANCOMPILE:
            self.assertLess(
                diff(iisignature.logsig(path1, s, "C"), blankLogSig), 0.000000001
            )
        self.assertLess(
            diff(iisignature.logsig(path1, s, "O"), blankLogSig), 0.000000001
        )
        self.assertLess(
            diff(iisignature.logsig(path1, s, "S"), blankLogSig), 0.000000001
        )
        self.assertLess(diff(iisignature.logsig(path1, s, "X"), blankSig), 0.000000001)
        self.assertLess(
            diff(iisignature.logsig(path1, s_a, "A"), blankLogSig_a), 0.000000001
        )
        blankLogSig[:d] = path2[1] - path2[0]
        blankLogSig_a[:d] = path2[1] - path2[0]
        blankSig[:d] = path2[1] - path2[0]
        if CANCOMPILE:
            self.assertLess(
                diff(iisignature.logsig(path2, s, "C"), blankLogSig), 0.000001
            )
        self.assertLess(diff(iisignature.logsig(path2, s, "O"), blankLogSig), 0.000001)
        self.assertLess(diff(iisignature.logsig(path2, s, "S"), blankLogSig), 0.000001)
        self.assertLess(diff(iisignature.logsig(path2, s, "X"), blankSig), 0.000001)
        self.assertLess(
            diff(iisignature.logsig(path2, s_a, "A"), blankLogSig_a), 0.000001
        )

    def testLevel1(self):
        m = 1
        d = 2
        path = numpy.random.uniform(size=(10, d))
        rightSig = path[-1, :] - path[0, :]
        s = prepare(d, m, "cosx2")
        self.assertLess(diff(iisignature.sig(path, m), rightSig), 0.0000001)
        for type_ in methodfilter(("C", "O", "S", "X", "A")):
            self.assertLess(
                diff(iisignature.logsig(path, s, type_), rightSig), 0.0000001, type_
            )
        self.assertLess(diff(rightSig, iisignature.logsigtosig(rightSig, s)), 0.000001)
        derivs = numpy.array([2.1, 3.2])
        pathderivs = numpy.zeros_like(path)
        pathderivs[-1] = derivs
        pathderivs[0] = -derivs
        self.assertLess(
            diff(iisignature.logsigbackprop(derivs, path, s), pathderivs), 0.00001
        )
        self.assertLess(
            diff(iisignature.logsigbackprop(derivs, path, s, "X"), pathderivs), 0.00001
        )
        self.assertLess(
            diff(iisignature.sigbackprop(derivs, path, m), pathderivs), 0.00001
        )

    def testHighDim(self):
        for m in [1, 2]:
            d = 1000
            path = numpy.random.rand(10, d)
            s = iisignature.prepare(d, m)
            iisignature.logsig(path, s, "A")
            # not testing result, just that it calculates something

    def testCumulative(self):
        m = 3
        d = 2
        length = 10
        path = numpy.random.uniform(size=(length, d))
        cumul = iisignature.sig(path, m, 2)
        expected = numpy.array(
            [iisignature.sig(path[: (i + 1)], m) for i in range(1, length)]
        )
        self.assertTrue(numpy.allclose(expected, cumul))
        path = numpy.random.uniform(size=(3, 2, length, d))
        cumul = iisignature.sig(path, m, 2)
        # expected = numpy.stack([iisignature.sig(path[:,:,:(i+1)],m) for i in range(1,length)],-2)
        expected = numpy.rollaxis(
            stack(
                [iisignature.sig(path[:, :, : (i + 1)], m) for i in range(1, length)]
            ),
            0,
            3,
        )
        self.assertTrue(numpy.allclose(expected, cumul))

    def testSquareLevel2(self):
        path = numpy.array([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]])
        m = 2
        sig = iisignature.sig(path, m)
        self.assertTrue(numpy.allclose(sig, [0, 0, 0, 1, -1, 0]))

    def testEllLevel2(self):
        path = numpy.array([[0, 0], [1, 0], [1, 1]])
        m = 2
        sig = iisignature.sig(path, m)
        self.assertTrue(numpy.allclose(sig, [1, 1, 0.5, 1, 0, 0.5]))

    def test1d(self):
        path = numpy.array([[0], [1], [3]])
        m = 3
        sig = iisignature.sig(path, m)
        self.assertTrue(numpy.allclose(sig, [3, 4.5, 4.5]))

    def test_big_siglength(self):
        self.assertEqual(iisignature.siglength(9, 9), sum(9**i for i in range(1, 10)))
        self.assertEqual(iisignature.siglength(9500, 2), 9500**2 + 9500)


class Scales(TestCase):
    # check sigscale and its derivatives
    def testa(self):
        numpy.random.seed(775)
        d = 3
        m = 5
        pathLength = 5
        numberToDo = 2
        paths = numpy.random.uniform(size=(numberToDo, pathLength, d))
        sigs = numpy.vstack([iisignature.sig(i, m) for i in paths])
        scales = numpy.random.uniform(0.5, 0.97, size=(numberToDo, d))
        scaledPaths = paths * scales[:, numpy.newaxis, :]
        scaledSigs = numpy.vstack([iisignature.sig(i, m) for i in scaledPaths])
        scaledSigsCalc = iisignature.sigscale(sigs, scales, m)
        self.assertEqual(scaledSigs.shape, scaledSigsCalc.shape)
        self.assertLess(diff(scaledSigs, scaledSigsCalc), 0.0000001)

        bumpedScales = 1.001 * scales
        bumpedSigs = 1.001 * sigs
        base = numpy.sum(scaledSigsCalc)
        bump1 = numpy.sum(iisignature.sigscale(bumpedSigs, scales, m))
        bump2 = numpy.sum(iisignature.sigscale(sigs, bumpedScales, m))
        derivsOfSum = numpy.ones_like(scaledSigsCalc)
        calculated = iisignature.sigscalebackprop(derivsOfSum, sigs, scales, m)
        diff1 = (bump1 - base) - numpy.sum(calculated[0] * (bumpedSigs - sigs))
        diff2 = (bump2 - base) - numpy.sum(calculated[1] * (bumpedScales - scales))
        # print(calculated[1].shape,bumpedScales.shape,scales.shape)
        # print(calculated[1][0,0],bump2,base,bumpedScales[0,0],scales[0,0])
        # print (bump1,bump2,base,diff1,diff2)
        self.assertLess(numpy.abs(diff1), 0.0000001)
        self.assertLess(numpy.abs(diff2), 0.0000001)


class Bases(TestCase):
    # Check that the Lyndon basis consists of Lyndon words.
    def testLyndon(self):
        d = 2
        m = 5
        s = iisignature.prepare(d, m, "O")
        for expression in iisignature.basis(s):
            word = "".join(c for c in expression if c not in "[,]")
            if len(word) > 1:
                for prefixLength in range(1, len(word)):
                    self.assertLess(word[:prefixLength], word[prefixLength:])


class Batching(TestCase):
    def test_batch(self):
        numpy.random.seed(734)
        d = 2
        m = 2
        n = 15
        paths = [numpy.random.uniform(-1, 1, size=(6, d)) for i in range(n)]
        pathArray15 = stack(paths)
        pathArray1315 = numpy.reshape(pathArray15, (1, 3, 1, 5, 6, d))
        sigs = [iisignature.sig(i, m) for i in paths]
        sigArray15 = iisignature.sig(pathArray15, m)
        sigArray1315 = iisignature.sig(pathArray1315, m)
        siglength = iisignature.siglength(d, m)
        self.assertEqual(sigArray1315.shape, (1, 3, 1, 5, siglength))
        self.assertTrue(numpy.allclose(sigArray1315.reshape(n, siglength), sigs))
        self.assertEqual(sigArray15.shape, (15, siglength))
        self.assertTrue(numpy.allclose(sigArray15, sigs))

        backsigs = [iisignature.sigbackprop(i, j, m) for i, j in zip(sigs, paths)]
        backsigArray = stack(backsigs)
        backsigs1315 = iisignature.sigbackprop(sigArray1315, pathArray1315, m)
        self.assertEqual(backsigs1315.shape, (1, 3, 1, 5, 6, d))
        self.assertTrue(numpy.allclose(backsigs1315.reshape(n, 6, 2), backsigArray))

        data = [numpy.random.uniform(size=(d,)) for i in range(n)]
        dataArray1315 = stack(data).reshape((1, 3, 1, 5, d))
        joined = [iisignature.sigjoin(i, j, m) for i, j in zip(sigs, data)]
        joined1315 = iisignature.sigjoin(sigArray1315, dataArray1315, m)
        self.assertEqual(joined1315.shape, (1, 3, 1, 5, siglength))
        self.assertTrue(numpy.allclose(joined1315.reshape(n, -1), stack(joined)))
        backjoined = [
            iisignature.sigjoinbackprop(i, j, k, m)
            for i, j, k in zip(joined, sigs, data)
        ]
        backjoinedArrays = [stack([i[j] for i in backjoined]) for j in range(2)]
        backjoined1315 = iisignature.sigjoinbackprop(
            joined1315, sigArray1315, dataArray1315, m
        )
        self.assertEqual(backjoined1315[0].shape, sigArray1315.shape)
        self.assertEqual(backjoined1315[1].shape, dataArray1315.shape)
        self.assertTrue(
            numpy.allclose(backjoined1315[0].reshape(n, -1), backjoinedArrays[0])
        )
        self.assertTrue(
            numpy.allclose(backjoined1315[1].reshape(n, -1), backjoinedArrays[1])
        )

        dataAsSigs = [
            iisignature.sig(numpy.vstack([numpy.zeros((d,)), i]), m) for i in data
        ]
        dataArray13151 = dataArray1315[:, :, :, :, None, :]
        dataArray13151 = numpy.repeat(dataArray13151, 2, 4) * [[0.0], [1.0]]
        dataArrayAsSigs1315 = iisignature.sig(dataArray13151, m)
        combined1315 = iisignature.sigcombine(sigArray1315, dataArrayAsSigs1315, d, m)
        self.assertEqual(joined1315.shape, combined1315.shape)
        self.assertTrue(numpy.allclose(joined1315, combined1315))
        backcombined1315 = iisignature.sigcombinebackprop(
            joined1315, sigArray1315, dataArrayAsSigs1315, d, m
        )
        backcombined = [
            iisignature.sigcombinebackprop(i, j, k, d, m)
            for i, j, k in zip(joined, sigs, dataAsSigs)
        ]
        backcombinedArrays = [stack([i[j] for i in backcombined]) for j in range(2)]
        self.assertEqual(backcombined1315[0].shape, sigArray1315.shape)
        self.assertEqual(backcombined1315[1].shape, sigArray1315.shape)
        self.assertTrue(numpy.allclose(backjoined1315[0], backcombined1315[0]))
        self.assertTrue(
            numpy.allclose(backcombined1315[0].reshape(n, -1), backcombinedArrays[0])
        )
        self.assertTrue(
            numpy.allclose(backcombined1315[1].reshape(n, -1), backcombinedArrays[1])
        )

        scaled = [iisignature.sigscale(i, j, m) for i, j in zip(sigs, data)]
        scaled1315 = iisignature.sigscale(sigArray1315, dataArray1315, m)
        self.assertEqual(scaled1315.shape, (1, 3, 1, 5, siglength))
        self.assertTrue(numpy.allclose(scaled1315.reshape(n, -1), stack(scaled)))
        backscaled = [
            iisignature.sigscalebackprop(i, j, k, m)
            for i, j, k in zip(scaled, sigs, data)
        ]
        backscaledArrays = [stack([i[j] for i in backscaled]) for j in range(2)]
        backscaled1315 = iisignature.sigscalebackprop(
            scaled1315, sigArray1315, dataArray1315, m
        )
        self.assertEqual(backscaled1315[0].shape, sigArray1315.shape)
        self.assertEqual(backscaled1315[1].shape, dataArray1315.shape)
        self.assertTrue(
            numpy.allclose(backscaled1315[0].reshape(n, -1), backscaledArrays[0])
        )
        self.assertTrue(
            numpy.allclose(backscaled1315[1].reshape(n, -1), backscaledArrays[1])
        )

        s_s = (prepare(d, m, "cosax2"), prepare(d, m, "cosahx2"))
        for type in ("c", "o", "s", "x", "a", "ch", "oh", "sh", "ah"):
            if type[0] == "c" and not CANCOMPILE:
                continue
            s = s_s[1 if "h" in type else 0]
            logsigs = [iisignature.logsig(i, s, type) for i in paths]
            logsigArray = stack(logsigs)
            logsigArray1315 = iisignature.logsig(pathArray1315, s, type)
            self.assertEqual(
                logsigArray1315.shape, (1, 3, 1, 5, logsigs[0].shape[0]), type
            )
            self.assertTrue(
                numpy.allclose(logsigArray1315.reshape(n, -1), logsigArray), type
            )

            if type in ("s", "x", "sh"):
                backlogs = stack(
                    [
                        iisignature.logsigbackprop(i, j, s, type)
                        for i, j in zip(logsigs, paths)
                    ]
                )
                backlogs1315 = iisignature.logsigbackprop(
                    logsigArray1315, pathArray1315, s, type
                )
                self.assertEqual(backlogs1315.shape, backsigs1315.shape)
                self.assertTrue(
                    numpy.allclose(backlogs1315.reshape(n, 6, d), backlogs), type
                )

        for s in s_s:
            logsigs1315 = iisignature.logsig(pathArray1315, s)
            logsigs = [iisignature.logsig(p, s) for p in paths]
            sigsFromLogSigs1315 = iisignature.logsigtosig(logsigs1315, s)
            self.assertEqual(sigsFromLogSigs1315.shape, sigArray1315.shape)
            self.assertTrue(numpy.allclose(sigsFromLogSigs1315, sigArray1315))

            backls2s = stack(
                [
                    iisignature.logsigtosigbackprop(i, j, s)
                    for i, j in zip(sigs, logsigs)
                ]
            )
            backls2s1315 = iisignature.logsigtosigbackprop(sigArray1315, logsigs1315, s)
            self.assertEqual(backls2s1315.shape, logsigs1315.shape)
            self.assertTrue(
                numpy.allclose(backls2s1315.reshape(-1, logsigs[0].shape[0]), backls2s)
            )

        a = iisignature.rotinv2dprepare(m, "a")
        rots = stack([iisignature.rotinv2d(i, a) for i in paths])
        rots1315 = iisignature.rotinv2d(pathArray1315, a)
        self.assertEqual(rots1315.shape, (1, 3, 1, 5, rots.shape[1]))
        self.assertTrue(numpy.allclose(rots1315.reshape(n, -1), rots))


# sum (2i choose i) for i in 1 to n
# which is the number of linear rotational invariants up to level 2n
def sumCentralBinomialCoefficient(n):
    f = math.factorial
    return sum(f(2 * i) / (f(i) ** 2) for i in range(1, n + 1))


# can run just this with
# python setup.py test -s tests.test_sig.RotInv2d
class RotInv2d(TestCase):
    def dotest(self, type):
        m = 8
        nPaths = 95
        nAngles = 348
        numpy.random.seed(775)
        s = iisignature.rotinv2dprepare(m, type)
        coeffs = iisignature.rotinv2dcoeffs(s)
        angles = numpy.random.uniform(0, math.pi * 2, size=nAngles + 1)
        angles[0] = 0
        rotationMatrices = [
            numpy.array([[math.cos(i), math.sin(i)], [-math.sin(i), math.cos(i)]])
            for i in angles
        ]
        paths = [numpy.random.uniform(-1, 1, size=(32, 2)) for i in range(nPaths)]
        samePathRotInvs = [
            iisignature.rotinv2d(numpy.dot(paths[0], mtx), s)
            for mtx in rotationMatrices
        ]

        # check the length matches
        (length,) = samePathRotInvs[0].shape
        self.assertEqual(length, sum(i.shape[0] for i in coeffs))
        self.assertEqual(length, iisignature.rotinv2dlength(s))
        if type == "a":
            self.assertEqual(length, sumCentralBinomialCoefficient(m // 2))

        self.assertLess(length, nAngles)  # sanity check on the test itself

        # check that the invariants are invariant
        if 0:
            print("\n", numpy.column_stack(samePathRotInvs[0:7]))
        for i in range(nAngles):
            if 0 and diff(samePathRotInvs[0], samePathRotInvs[1 + i]) > 0.01:
                print(i)
                print(samePathRotInvs[0] - samePathRotInvs[1 + i])
                print(diff(samePathRotInvs[0], samePathRotInvs[1 + i]))
            self.assertLess(diff(samePathRotInvs[0], samePathRotInvs[1 + i]), 0.01)

        # check that the invariants match the coefficients
        if 1:
            sigLevel = iisignature.sig(paths[0], m)[iisignature.siglength(2, m - 1) :]
            lowerRotinvs = (
                0
                if 2 == m
                else iisignature.rotinv2dlength(
                    iisignature.rotinv2dprepare(m - 2, type)
                )
            )
            # print("\n",numpy.dot(coeffs[-1],sigLevel),"\n",samePathRotInvs[0][lowerRotinvs:])
            # print(numpy.dot(coeffs[-1],sigLevel)-samePathRotInvs[0][lowerRotinvs:])
            self.assertTrue(
                numpy.allclose(
                    numpy.dot(coeffs[-1], sigLevel),
                    samePathRotInvs[0][lowerRotinvs:],
                    atol=0.000001,
                )
            )

        # check that we are not missing invariants
        if type == "a":
            # print("\nrotinvlength=",length,"
            # siglength=",iisignature.siglength(2,m))
            sigOffsets = []
            for path in paths:
                samePathSigs = [
                    iisignature.sig(numpy.dot(path, mtx), m)
                    for mtx in rotationMatrices[1:70]
                ]
                samePathSigsOffsets = [i - samePathSigs[0] for i in samePathSigs[1:]]
                sigOffsets.extend(samePathSigsOffsets)

            # print(numpy.linalg.svd(numpy.vstack(sigOffsets))[1])
            def split(a, dim, level):
                start = 0
                out = []
                for m in range(1, level + 1):
                    levelLength = dim**m
                    out.append(a[:, start : (start + levelLength)])
                    start = start + levelLength
                assert start == a.shape[1]
                return out

            allOffsets = numpy.vstack(sigOffsets)
            # print (allOffsets.shape)
            splits = split(allOffsets, 2, m)
            # print()
            rank_tolerance = 0.01  # this is hackish
            # print
            # ([numpy.linalg.matrix_rank(i.astype("float64"),rank_tolerance)
            # for
            # i in splits])
            # print ([i.shape for i in splits])
            # print(numpy.linalg.svd(splits[-1])[1])

            # sanity check on the test
            self.assertLess(splits[-1].shape[1], splits[0].shape[0])
            totalUnspannedDimensions = sum(
                i.shape[1] - numpy.linalg.matrix_rank(i, rank_tolerance) for i in splits
            )
            self.assertEqual(totalUnspannedDimensions, length)

        if 0:  # This doesn't work - the rank of the whole thing is less than
            # sigLength-totalUnspannedDimensions, which suggests that there are
            # inter-level dependencies,
            # even though the shuffle product dependencies aren't linear.
            # I don't know why this is.
            sigLength = iisignature.siglength(2, m)
            numNonInvariant = numpy.linalg.matrix_rank(numpy.vstack(sigOffsets))

            predictedNumberInvariant = sigLength - numNonInvariant
            print(sigLength, length, numNonInvariant)
            self.assertLess(sigLength, nAngles)
            self.assertEqual(predictedNumberInvariant, length)

    def test_a(self):
        self.dotest("a")

    def test_k(self):
        self.dotest("k")

    def test_q(self):
        if HAVE_SCIPY:
            self.dotest("q")

    def test_s(self):
        self.dotest("s")

    def testConsistencyOfBases(self):
        m = 6
        sa = iisignature.rotinv2dprepare(m, "a")
        sk = iisignature.rotinv2dprepare(m, "k")
        ca = iisignature.rotinv2dcoeffs(sa)[-1]
        ck = iisignature.rotinv2dcoeffs(sk)[-1]

        # every row of ck should be in the span of the rows of ca
        # i.e.  every column of ck.T should be in the span of the columns of
        # ca.T
        # i.e.  there's a matrix b s.t.  ca.T b = ck.T
        residuals = lstsq(ca.T, ck.T)[1]
        self.assertLess(numpy.max(numpy.abs(residuals)), 0.000001)

        ss = iisignature.rotinv2dprepare(m, "s")
        cs = iisignature.rotinv2dcoeffs(ss)[-1]
        # every row of cs and cq should be in the span of the rows of ca
        residuals2 = lstsq(ca.T, cs.T)[1]
        self.assertLess(numpy.max(numpy.abs(residuals2)), 0.000001)
        if HAVE_SCIPY:
            sq = iisignature.rotinv2dprepare(m, "q")
            cq = iisignature.rotinv2dcoeffs(sq)[-1]
            residuals2 = lstsq(ca.T, cq.T)[1]
            self.assertLess(numpy.max(numpy.abs(residuals2)), 0.000001)

            self.assertEqual(cq.shape, cs.shape)
        else:
            cq = None

        # check that the invariants are linearly independent (not for k)
        for c, name in ((cs, "s"), (ca, "a"), (cq, "q")):
            if not HAVE_SCIPY and name == "q":
                continue
            self.assertEqual(numpy.linalg.matrix_rank(c), c.shape[0], name)

        # check that rows with nonzeros in evil columns are all before
        # rows with nonzeros in odious columns
        # print ((numpy.abs(ca)>0.00000001).astype("int8"))
        for c, name in ((cs, "s"), (ck, "k"), (ca, "a"), (cq, "q")):
            if not HAVE_SCIPY and name == "q":
                continue
            evilRows = []
            odiousRows = []
            for i in range(c.shape[0]):
                evil = 0
                odious = 0
                for j in range(c.shape[1]):
                    if numpy.abs(c[i, j]) > 0.00001:
                        if bin(j).count("1") % 2:
                            odious = odious + 1
                        else:
                            evil = evil + 1
                if evil > 0:
                    evilRows.append(i)
                if odious > 0:
                    odiousRows.append(i)
            # print (evilRows, odiousRows)
            self.assertLess(
                numpy.max(evil), numpy.min(odious), "bad order of rows in " + name
            )


class NonContiguous(unittest.TestCase):
    """Test that sig (and friends) produce correct results for non-contiguous inputs."""

    def test_sig_noncontiguous_batch(self):
        numpy.random.seed(891)
        paths = numpy.random.randn(10, 20, 3)
        m = 4
        ref = iisignature.sig(paths[::2].copy(), m)
        got = iisignature.sig(paths[::2], m)
        self.assertTrue(numpy.allclose(got, ref))

    def test_sig_fortran_order(self):
        numpy.random.seed(892)
        paths = numpy.random.randn(10, 20, 3)
        m = 4
        ref = iisignature.sig(paths, m)
        got = iisignature.sig(numpy.asfortranarray(paths), m)
        self.assertTrue(numpy.allclose(got, ref))

    def test_sig_noncontiguous_batch_multidim(self):
        numpy.random.seed(893)
        paths = numpy.random.randn(6, 4, 8, 3)
        m = 3
        ref = iisignature.sig(paths[::2, ::2].copy(), m)
        got = iisignature.sig(paths[::2, ::2], m)
        self.assertTrue(numpy.allclose(got, ref))

    def test_sig_cumulative_noncontiguous(self):
        numpy.random.seed(894)
        paths = numpy.random.randn(10, 20, 3)
        m = 3
        ref = iisignature.sig(paths[::2].copy(), m, 2)
        got = iisignature.sig(paths[::2], m, 2)
        self.assertTrue(numpy.allclose(got, ref))


class Float32Input(unittest.TestCase):
    """Test that float32 path inputs produce correct results (matching float64)."""

    def test_sig_float32(self):
        numpy.random.seed(901)
        path = numpy.random.uniform(size=(10, 3)).astype(numpy.float32)
        result_f32 = iisignature.sig(path, 3)
        result_f64 = iisignature.sig(path.astype(numpy.float64), 3)
        numpy.testing.assert_allclose(result_f32, result_f64, rtol=1e-5)

    def test_sig_float32_fortran_order(self):
        numpy.random.seed(902)
        path = numpy.asfortranarray(
            numpy.random.uniform(size=(10, 3)).astype(numpy.float32)
        )
        result = iisignature.sig(path, 3)
        expected = iisignature.sig(
            numpy.ascontiguousarray(path).astype(numpy.float64), 3
        )
        numpy.testing.assert_allclose(result, expected, rtol=1e-5)

    def test_sig_float32_batch(self):
        numpy.random.seed(903)
        paths = numpy.random.uniform(size=(5, 10, 3)).astype(numpy.float32)
        result_f32 = iisignature.sig(paths, 3)
        result_f64 = iisignature.sig(paths.astype(numpy.float64), 3)
        numpy.testing.assert_allclose(result_f32, result_f64, rtol=1e-5)

    def test_sig_float32_cumulative(self):
        numpy.random.seed(904)
        path = numpy.random.uniform(size=(10, 3)).astype(numpy.float32)
        result_f32 = iisignature.sig(path, 3, 2)
        result_f64 = iisignature.sig(path.astype(numpy.float64), 3, 2)
        numpy.testing.assert_allclose(result_f32, result_f64, rtol=1e-5)

    def test_sigbackprop_float32(self):
        numpy.random.seed(905)
        d = 3
        m = 3
        path = numpy.random.uniform(size=(10, d)).astype(numpy.float32)
        sig = iisignature.sig(path, m)
        derivs = numpy.ones_like(sig)
        result_f32 = iisignature.sigbackprop(derivs, path, m)
        result_f64 = iisignature.sigbackprop(derivs, path.astype(numpy.float64), m)
        numpy.testing.assert_allclose(result_f32, result_f64, rtol=1e-5)

    def test_logsig_float32_all_methods(self):
        numpy.random.seed(906)
        d = 3
        path = numpy.random.uniform(size=(10, d)).astype(numpy.float32)
        path_f64 = path.astype(numpy.float64)
        for m, methods in [(2, "COSA"), (3, "COS")]:
            for method in methodfilter(list(methods)):
                s = prepare(d, m, method.lower())
                result_f32 = iisignature.logsig(path, s, method)
                result_f64 = iisignature.logsig(path_f64, s, method)
                numpy.testing.assert_allclose(
                    result_f32,
                    result_f64,
                    rtol=1e-5,
                    err_msg="logsig method %s level %d" % (method, m),
                )

    def test_logsig_float32_expanded(self):
        numpy.random.seed(907)
        d = 3
        m = 3
        path = numpy.random.uniform(size=(10, d)).astype(numpy.float32)
        s = prepare(d, m, "x")
        result_f32 = iisignature.logsig(path, s, "X")
        result_f64 = iisignature.logsig(path.astype(numpy.float64), s, "X")
        numpy.testing.assert_allclose(result_f32, result_f64, rtol=1e-5, atol=1e-15)

    def test_logsigbackprop_float32(self):
        numpy.random.seed(908)
        d = 3
        m = 2
        path = numpy.random.uniform(size=(10, d)).astype(numpy.float32)
        for method in methodfilter(["S", "X", "A"]):
            s = prepare(d, m, method.lower())
            logsig = iisignature.logsig(path, s, method)
            derivs = numpy.ones_like(logsig)
            result_f32 = iisignature.logsigbackprop(derivs, path, s, method)
            result_f64 = iisignature.logsigbackprop(
                derivs, path.astype(numpy.float64), s, method
            )
            numpy.testing.assert_allclose(
                result_f32,
                result_f64,
                rtol=1e-5,
                err_msg="logsigbackprop method %s" % method,
            )

    def test_rotinv2d_float32(self):
        numpy.random.seed(909)
        path = numpy.random.uniform(size=(12, 2)).astype(numpy.float32)
        s = iisignature.rotinv2dprepare(4, "a")
        result_f32 = iisignature.rotinv2d(path, s)
        result_f64 = iisignature.rotinv2d(path.astype(numpy.float64), s)
        numpy.testing.assert_allclose(result_f32, result_f64, rtol=1e-5)


if __name__ == "__main__":
    sys.path.append("..")
    import iisignature

    # This is convenient for running some tests just by running this file,
    # but only works with python3 (you may need to have built inplace first)
    a = Scales().testa()



# A trivial demonstration of the RecurrentSig layer from iisignature_recurrent_torch.py
# No assertion is made that this model is a good idea, or that this code is idiomatic pytorch.

import numpy as np
import sys
import os
import itertools
import torch
from torch.autograd import Variable
import torch.nn as nn

# add the parent directory, so we find our iisignature build if it was built --inplace
sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))

from iisignature_recurrent_torch import RecurrentSig

criterion = nn.MSELoss()

# The task here for the network to learn is very easy - the average of two of the inputs
x = np.random.uniform(size=(2311, 5, 3))
y = (
    (x[:, 1, 1] + x[:, 3, 2]) / 2
)  # The output is a number between 0 and 1, so matches sigmoid activation of the top layer
testx = x[2000:, :, :]
testy = y[2000:]
x = x[:2000, :, :]
y = y[:2000]

rnn = RecurrentSig(
    3,
    5,
    sig_level=2,
    use_signatures=False,
    output_signatures=False,
    train_time_lapse=False,
)
finalLayer = nn.Linear(5, 1)

optimizer = torch.optim.Adam(
    itertools.chain(rnn.parameters(), finalLayer.parameters()), lr=0.0001
)

minibatch_size = 32


def train(x_batch, y_batch):
    minibatch_size = y_batch.shape[0]
    hidden = rnn.initHidden(minibatch_size)
    optimizer.zero_grad()
    x_batch = Variable(torch.FloatTensor(x_batch))
    for i in range(5):
        output, hidden = rnn(x_batch[:, i, :], hidden)

    output = finalLayer(output)
    loss = criterion(output, Variable(torch.FloatTensor(y_batch)))
    loss.backward()
    optimizer.step()
    return output, loss.data[0]


def predict(x):
    hidden = rnn.initHidden(x.shape[0])
    x = Variable(torch.FloatTensor(x))
    for i in range(5):
        output, hidden = rnn(x[:, i, :], hidden)
    output = finalLayer(output)
    return output


def evaluate(x, y):
    loss = criterion(predict(x), Variable(torch.FloatTensor(y)))
    return loss


nb_epoch = 2

for i in range(nb_epoch * x.shape[0] // minibatch_size):
    indices = np.random.randint(testx.shape[0], size=minibatch_size)
    output, loss = train(x[indices], y[indices])
    # print (loss)

# a=np.random.uniform(size=(3,5,3))
# print (a, predict(a).data)

print(evaluate(testx, testy).data)



import torch
from torch.autograd import Variable
# from torch.nn.modules.module import Module

import os
import sys

sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
import iisignature

from iisignature_torch import Sig, LogSig, SigJoin, SigScale


def trySig():
    inp = Variable(torch.randn(8, 2), requires_grad=True)
    result = Sig(inp, 2)
    print(result.data)
    result.backward(torch.randn(result.size()))
    print(inp.grad)


trySig()


def tryLogSig():
    s = iisignature.prepare(2, 2)
    for expanded in (False, True):
        inp = Variable(torch.randn(8, 2), requires_grad=True)
        if expanded:
            result = LogSig(inp, s, "x")
        else:
            result = LogSig(inp, s)
        print(result.data.numpy())
        result.backward(torch.randn(result.size()))
        print(inp.grad.data)


tryLogSig()


def trySigScale():
    inp1 = Variable(torch.randn(12, 6), requires_grad=True)
    inp2 = Variable(torch.randn(12, 2), requires_grad=True)
    result = SigScale(inp1, inp2, 2)
    print(result.data)
    result.backward(torch.randn(result.size()))
    print(inp1.grad)
    print(inp2.grad)


trySigScale()


def trySigJoin():
    inp1 = Variable(torch.randn(12, 6), requires_grad=True)
    inp2 = Variable(torch.randn(12, 2), requires_grad=True)
    result = SigJoin(inp1, inp2, 2)
    print(result.data)
    result.backward(torch.randn(result.size()))
    print(inp1.grad)
    print(inp2.grad)


trySigJoin()


def trySigJoin2():
    inp1 = Variable(torch.randn(12, 12), requires_grad=True)
    inp2 = Variable(torch.randn(12, 2), requires_grad=True)
    inp3 = Variable(torch.randn(1), requires_grad=True)
    result = SigJoin(inp1, inp2, 2, inp3)
    print(result.data)
    result.backward(torch.randn(result.size()))
    print(inp1.grad)
    print(inp2.grad)
    print(inp3.grad)


trySigJoin2()



import keras.layers.recurrent
from keras.engine import InputSpec
from keras import backend as K
from keras import initializers, activations
import keras
import iisignature

if K.backend() == "theano":
    from iisignature_theano import SigJoin
if K.backend() == "tensorflow":
    from iisignature_tensorflow import SigJoin
import math
from distutils.version import LooseVersion

if LooseVersion(keras.__version__) < LooseVersion("2.0.4"):
    raise RuntimeError("old keras, please find another version of this file")

# Consider initializing so that state is constant initially or something.
# -perhaps a highway
# add dropout, regularizers


class RecurrentSig(keras.layers.recurrent.Recurrent):
    """
    A recurrent layer like keras's SimpleRNN or LSTM, which includes the signatures of the graph
    of each state against time.

    n_units: the number of units, this is not the number of outputs if output_signatures is Triue

    sig_level: the depth of the signature

    initial_time_lapse: the distance in the time dimension between previous values when
    calculating a 2d signature

    train_time_lapse: whether to train the time lapse, this doesn't work yet because
    SigJoinGrad doesn't return a gradient wrt the time lapse.

    output_signatures: whether the signatures as well as the current hidden state are
    output to succeeding layers

    use_signatures: whether the new state can depend on the value of the signature.
    Note that this layer is just like a simple RNN if use_signatures
    and output_signatures are both false.

    kernel_initializer: the initialisation of the map from input to state

    recurrent_initializer: the initialisation of the matrix from state to state

    activation: the activation applied to the output values of the state.
    """

    def __init__(
        self,
        n_units,
        sig_level=2,
        train_time_lapse=False,
        initial_time_lapse=0.1,
        output_signatures=False,  # whether the output includes the signatures
        use_signatures=True,  # whether each new state can depend on the signature
        kernel_initializer="glorot_uniform",
        recurrent_initializer="he_normal",  # could be good to use 'orthogonal' if not use_signatures
        activation="tanh",  # not applied to signature elements
        **kwargs,
    ):
        self.sig_level = sig_level
        self.sigsize = iisignature.siglength(2, sig_level)
        self.n_units = n_units  # like output_dim
        self.units = n_units * (self.sigsize + 1) if output_signatures else n_units
        self.train_time_lapse = train_time_lapse
        self.initial_time_lapse = initial_time_lapse
        self.output_signatures = output_signatures
        self.use_signatures = use_signatures
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.recurrent_initializer = initializers.get(recurrent_initializer)
        self.activation = activations.get(activation)
        super(RecurrentSig, self).__init__(**kwargs)

    def build(self, input_shape):
        self.input_spec = [InputSpec(shape=input_shape)]
        self.input_dim = input_shape[2]
        self.states = [None, None]  # ?
        self.W = self.add_weight(
            shape=(self.input_dim, self.n_units),
            name="W",
            initializer=self.kernel_initializer,
        )
        saved_data_length = self.n_units * (
            1 + self.sigsize if self.use_signatures else 1
        )
        self.U = self.add_weight(
            shape=(saved_data_length, self.n_units),
            name="U",
            initializer=self.recurrent_initializer,
        )
        self.b = K.zeros((self.n_units,), name="{}_b".format(self.name))

        self.trainable_weights = [self.W, self.U, self.b]
        if self.train_time_lapse:
            log_timelapse = K.variable(
                math.log(self.initial_time_lapse),
                name="{}_log_time_lapse".format(self.name),
            )
            self.trainable_weights.append(log_timelapse)
            self.time_lapse = K.exp(log_timelapse)
        else:
            self.time_lapse = self.initial_time_lapse

    def step(self, x, states):
        prev_sigs = states[0]  # (batch,n_units*sigsize)
        prev_sigs_ = K.reshape(prev_sigs, (-1, self.sigsize))  # (batch*n_units,sigsize)
        prev_states = states[1]
        h = K.dot(x, self.W) + self.b
        prev_states_as_memory = (
            K.concatenate([prev_sigs, prev_states], 1)
            if self.use_signatures
            else prev_states
        )
        raw_output = self.activation(h + K.dot(prev_states_as_memory, self.U))
        displacements = K.reshape(raw_output - prev_states, (-1, 1))
        sigs = SigJoin(prev_sigs_, displacements, self.sig_level, self.time_lapse)
        sigs = K.reshape(sigs, (-1, self.n_units * self.sigsize))
        activated_output_units = self.activation(raw_output)
        if self.output_signatures:
            output = K.concatenate([sigs, activated_output_units], 1)
        else:
            output = activated_output_units
        return output, [sigs, raw_output]

    def get_initial_state(self, x):
        # x has shape (samples, timesteps, input_dim)
        # build all-zero tensors of shape (samples, whatever)
        initial_state = K.zeros_like(x)
        initial_state = K.sum(initial_state, axis=(1, 2))  # (samples,)
        initial_state = K.expand_dims(initial_state)  # (samples, 1)
        lengths = (self.n_units * self.sigsize, self.n_units)
        initial_states = [
            K.tile(initial_state, [1, i]) for i in lengths
        ]  # (samples, i)
        return initial_states

    def get_config(self):
        config = {
            "sigsize": self.sigsize,
            "sig_level": self.sig_level,
            "train_time_lapse": self.train_time_lapse,
            "initial_time_lapse": self.initial_time_lapse,
            "n_units": self.n_units,
            "units": self.units,
            "output_signatures": self.output_signatures,
            "use_signatures": self.use_signatures,
            "kernel_initializer": initializers.serialize(self.kernel_initializer),
            "recurrent_initializer": initializers.serialize(self.recurrent_initializer),
            "activation": self.activation.__name__,
        }
        base_config = super(RecurrentSig, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



# This boring example just shows that using SigJoin and Sig
# you get the same signatures and derivatives.

import os

# os.environ["THEANO_FLAGS"]="floatX=float32,device=cpu,optimizer=fast_compile"
# os.environ["THEANO_FLAGS"]="floatX=float32,device=cpu,mode=DebugMode"
os.environ["THEANO_FLAGS"] = "floatX=float32,device=cpu"

import theano
import numpy
import sys
import six.moves

# add the parent directory, so we find our iisignature build if it was built --inplace
sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
import iisignature

from iisignature_theano import Sig, SigJoin
import theano.tensor as T


# 1: SETUP VARIABLES
dim = 3
level = 6
pathlength = 4
fixed = float("nan")
fixed = 0.1

numpy.random.seed(51)
start = numpy.random.uniform(size=(pathlength, dim)).astype("float32")

# 2: DEFINE THEANO STUFF

path = theano.shared(start, "path")
if not numpy.isnan(fixed):
    fixedT = theano.shared(fixed, "fixedT")
    path = theano.tensor.horizontal_stack(
        path[:, :-1], fixedT * T.shape_padright(T.arange(pathlength))
    )
cost1 = theano.tensor.mean(theano.tensor.sqr(Sig(path, level)))
grad1 = theano.grad(cost1, path)

signature = numpy.zeros((1, iisignature.siglength(dim, level))).astype("float32")
for i in six.moves.xrange(1, pathlength):
    if not numpy.isnan(fixed):
        displacement = path[i : (i + 1), :-1] - path[(i - 1) : i, :-1]
        signature = SigJoin(signature, displacement, level, fixedT)
    else:
        displacement = path[i : (i + 1), :] - path[(i - 1) : i, :]
        signature = SigJoin(signature, displacement, level)

cost2 = theano.tensor.mean(theano.tensor.sqr(signature))
grad2 = theano.grad(cost2, path)

# theano.printing.pydotprint(grad2,outfile="a.png")

ff = theano.function([], [grad1, grad2])
# theano.printing.pydotprint(ff,outfile="b.png")

# 3: GO

numpy.set_printoptions(suppress=True)
f = ff()
if numpy.isnan(fixed):
    print(f[0] - f[1])
    print(f[0])
    print(f[1])
else:
    print((f[0] - f[1])[:, :-1])
    print(f[0][:, :-1])
    print(f[1][:, :-1])
    ff2 = theano.function(
        [], [theano.grad(cost1, fixedT), theano.grad(cost2, fixedT)]
    )()
    print(ff2)



import subprocess
import sqlite3

# valgrind -q --tool=massif --massif-out-file=mem_profile.massif --depth=1 ./mem_profilee 2 8 S H
##ms_print massif.out.`cat mem_profile_pid.txt`
##ms_print mem_profile.massif
# grep mem_heap_B mem_profile.massif |cut -c 12- |sort -n | tail -n 1


def get_usage(d, m, use_compiled, use_lyndon):
    args = [
        "valgrind",
        "-q",
        "--tool=massif",
        "--massif-out-file=mem_profile.massif",
        "--depth=1",
        "./mem_profilee",
        str(d),
        str(m),
        "C" if use_compiled else "S",
        "L" if use_lyndon else "H",
    ]
    b = subprocess.check_output(args)
    extra_memory = int(b) if (use_compiled and d > 0) else 0
    a = subprocess.check_output(
        "grep mem_heap_B mem_profile.massif |cut -c 12- |sort -n | tail -n 1",
        shell=True,
    )
    return int(a) + extra_memory


con = sqlite3.connect("mem_profile.sqlite")
con.cursor().execute(
    "create table if not exists A (METHOD TEXT, BASIS TEXT, D INT, M INT, BYTES INT, PERFORMED TEXT, VERSION TEXT)"
)
con.commit()
version = 1


def test(d, m, use_compiled, use_lyndon, no_commit):
    usage = get_usage(d, m, use_compiled, use_lyndon)
    method = "C" if use_compiled else "S"
    basis = "L" if use_lyndon else "H"
    print(method, basis, d, m, usage)
    con.cursor().execute(
        "INSERT INTO A(METHOD,BASIS,D,M,BYTES,PERFORMED,VERSION) VALUES(?,?,?,?,?,CURRENT_TIMESTAMP,?)",
        (method, basis, d, m, usage, version),
    )
    if not no_commit:
        con.commit()


no_commit = 0
if no_commit:
    print("warning: not saving results")

# test(0,0,True,1,no_commit)
for use_lyndon in [True]:  # ,False]:
    for use_compiled in [True, False]:
        for d in [3]:
            for m in range(10, 11):
                test(d, m, use_compiled, use_lyndon, no_commit)



import tensorflow as tf

import os
import sys

sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
import iisignature

from iisignature_tensorflow import Sig, LogSig


def trySig():
    sess = tf.Session()
    a = tf.placeholder(tf.float32)
    b = tf.placeholder(tf.float32)
    out = Sig(a, b)
    grad = tf.gradients(out, a)[0]

    path = [[2, 3], [3, 5], [4, 5]]
    m = 2

    out_vals = sess.run([out, grad], {a: path, b: m})
    print(out_vals[0])
    print(out_vals[1])


# trySig()


def tryLogSig():
    for expanded in (False, True):
        sess = tf.Session()
        path = [[2, 3], [3, 5], [4, 5]]
        m = 2
        s = iisignature.prepare(2, m)
        a = tf.placeholder(tf.float32)
        out = LogSig(a, s, "x") if expanded else LogSig(a, s)
        grad = tf.gradients(tf.reduce_sum(out), a)[0]
        # grad = tf.gradients(out,a)[0]

        out_vals = sess.run([out, grad], {a: path})
        print(out_vals[0])
        print(out_vals[1])


tryLogSig()



import os
import sys
import numpy as np
from six.moves import range

# This program just continually runs some of the important functionality in
# iisignature on small data, so that you can check that memory usage
# does not grow. Even a tiny memory leak becomes very visible e.g. in htop.

# add the parent directory, so we find our iisignature build if it was built --inplace
sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
import iisignature

length = 20
dim = 3
level = 2
npaths = 3
paths_ = np.random.uniform(size=(npaths, length, dim))
scale_ = np.random.uniform(size=(npaths, dim))
initialsigs_ = np.random.uniform(size=(npaths, iisignature.siglength(dim, level)))
p = iisignature.prepare(dim, level, "cosx")
while 0:
    iisignature.sig(paths_[0], level)
for i in range(10**10):
    # copy major parts of the input data, in case we are leaking references to it
    paths = paths_[:]
    increment = scale = scale_[:]
    initialsigs = initialsigs_[:]
    iisignature.sigjoin(initialsigs, scale, level)
    iisignature.sigscale(initialsigs, scale, level)
    iisignature.sigjoinbackprop(initialsigs, initialsigs, scale, level)
    iisignature.sigscalebackprop(initialsigs, initialsigs, scale, level)
    iisignature.sig(paths[0, :, :], level)
    iisignature.sigbackprop(initialsigs[0, :], paths[0, :, :], level)
    # iisignature.sigjacobian(paths[0,:,:],level)
    # iisignature.prepare(dim,level,"cosx")#much slower than other functions
    iisignature.logsig(paths[0, :, :], p, "c")
    iisignature.logsig(paths[0, :, :], p, "o")
    iisignature.logsig(paths[0, :, :], p, "s")
    if i % 10000 == 0:
        print(i)



# This file provides a wrapper around iisignature which provides the
# same interface as esig.tosig.

import iisignature
import numpy as np

_prep = {}
_prepLyndon = {}


def _getPrep(d, m, lyndon=False):
    hash2use = _prepLyndon if lyndon else _prep
    key = (d, m)
    if key in hash2use:
        return hash2use[key]
    ans = iisignature.prepare(d, m, "D" if lyndon else "DH")
    hash2use[key] = ans
    return ans


def sigdim(d, m):
    return 1 + iisignature.siglength(d, m)


logsigdim = iisignature.logsiglength


def sigkeys(d, m):
    from itertools import chain, product

    alphabet = "123456789"[:d]
    it = chain.from_iterable(product(alphabet, repeat=r) for r in range(m + 1))
    return " " + " ".join("(" + ",".join(j for j in i) + ")" for i in it)


def logsigkeys(d, m):
    s = _getPrep(d, m)
    return " " + " ".join(iisignature.basis(s))


def stream2logsig(path, m):
    d = path.shape[1]
    s = _getPrep(d, m)
    return iisignature.logsig(path, s).astype("float64")


def stream2logsig_Lyndon(path, m):
    d = path.shape[1]
    s = _getPrep(d, m, True)
    return iisignature.logsig(path, s).astype("float64")


def stream2sig(path, m):
    return np.hstack([1.0, iisignature.sig(path, m)])



# A trivial demonstration of the RecurrentSig layer from iisignature_recurrent_keras.py
# relies on keras 2

import os

# os.environ["THEANO_FLAGS"]="floatX=float32,device=cpu,optimizer=fast_compile"
# os.environ["THEANO_FLAGS"]="floatX=float32,device=cpu,mode=DebugMode"
# os.environ["THEANO_FLAGS"]="floatX=float32,device=gpu0,force_device=True,cxx=g++-4.8,nvcc.flags=-D_FORCE_INLINES,nvcc.compiler_bindir=/usr/bin/g++-4.8"
# os.environ["THEANO_FLAGS"]="floatX=float32,device=gpu0,force_device=True,cxx=g++-4.8,nvcc.flags=-D_FORCE_INLINES,nvcc.compiler_bindir=/usr/bin/g++-4.8,base_compiledir=/run/user/1001/theano"
os.environ["THEANO_FLAGS"] = (
    "floatX=float32,device=cpu,force_device=True,mode=DebugMode,DebugMode.check_finite=False"
)
os.environ["THEANO_FLAGS"] = "floatX=float32,device=cpu,force_device=True"
# os.environ["THEANO_FLAGS"]="floatX=float32,device=cpu,force_device=True,mode=NanGuardMode,exception_verbosity=high,NanGuardMode.inf_is_error=False,NanGuardMode.big_is_error=False,NanGuardMode.action=warn,optimizer=fast_compile"
os.environ["KERAS_BACKEND"] = "theano"
os.environ["KERAS_BACKEND"] = "tensorflow"

import numpy
import sys

# add the parent directory, so we find our iisignature build if it was built --inplace
sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))

from iisignature_recurrent_keras import RecurrentSig
import keras.models
import keras.layers.recurrent
import keras.layers.core


m = keras.models.Sequential()
# a few possible networks here.
# using relu with RecurrentSig sometimes gets Nans

m.add(
    RecurrentSig(
        5,
        sig_level=2,
        input_shape=(None, 3),
        return_sequences=False,
        use_signatures=True,
        output_signatures=False,
        activation="tanh",
        train_time_lapse=True,
    )
)

# m.add(RecurrentSig(5,input_shape=(5,3),return_sequences=True, use_signatures = True, output_signatures = False, activation="relu"))
# m.add(RecurrentSig(6,return_sequences=False,activation="relu"))

# m.add(LSTM(5,input_shape=(5,3),return_sequences=False))

# m.add(LSTM(5,input_shape=(5,3),return_sequences=True))
# m.add(LSTM(6,return_sequences=False))

# m.add(keras.layers.core.Flatten(input_shape=(5,3)))
# m.add(keras.layers.core.Dense(1000,activation="relu"))

m.add(keras.layers.core.Dense(1, activation="sigmoid"))
op = keras.optimizers.Adam(lr=0.0002, beta_1=0.9, beta_2=0.999, epsilon=1e-8)
m.compile(loss="mse", optimizer=op)  # metrics = accuracy
m.summary()

# The task here for the network to learn is very easy - the average of two of the inputs
x = numpy.random.uniform(size=(2311, 5, 3))
y = (
    (x[:, 1, 1] + x[:, 3, 2]) / 2
)  # The output is a number between 0 and 1, so matches sigmoid activation of the top layer
testx = x[2000:, :, :]
testy = y[2000:]
x = x[:2000, :, :]
y = y[:2000]

# a=numpy.random.uniform(size=(3,5,3))
# print (m.predict(a))
m.fit(x, y, epochs=10, shuffle=0)
print(m.evaluate(testx, testy, verbose=0))



# This example shows a very simple Keras model which does nothing interesting,
# but has a layer which uses Sig from iisignature.
# It trains, giving some confidence in the derivatives.

import os
import sys

# os.environ["KERAS_BACKEND"]="tensorflow"
# os.environ["THEANO_FLAGS"]="mode=DebugMode,device=cpu,optimizer_excluding=local_shape_to_shape_i"
import numpy as np
import keras
import keras.models
import keras.layers.recurrent
import keras.layers.core
import keras.backend as K

sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
if K.backend() == "theano":
    from iisignature_theano import Sig
if K.backend() == "tensorflow":
    from iisignature_tensorflow import Sig
import iisignature

sig_level = 2
siglen = iisignature.siglength(2, sig_level)
nSamples = 21
xDim = 55

X = np.zeros((nSamples, xDim), dtype="float32")
Y = np.ones((nSamples,), dtype="float32")

m = keras.models.Sequential()

m.add(keras.layers.core.Dense(106, input_shape=(xDim,)))
m.add(keras.layers.core.Reshape((53, 2)))
m.add(keras.layers.core.Lambda(lambda x: Sig(x, 2), output_shape=(siglen,)))
m.add(keras.layers.core.Reshape((siglen, 1)))
m.add(keras.layers.pooling.AveragePooling1D(siglen))
m.add(keras.layers.core.Flatten())
m.compile(loss="mse", optimizer="sgd")
m.summary()

m.fit(X, Y, epochs=1000)

# https://github.com/tensorflow/tensorflow/issues/3388
if K.backend() == "tensorflow":
    K.clear_session()



# This file provides a function getMatrices which returns the dot products of logsignature
# basis elements as a matrix for each level.
# This allows you to find the L2 dot product of log signatures in tensor space without
# ever expressing them in tensor space.
# This might be useful if you have to calculate such a dot product very many times.
# The demo() function provides an example.

# Don't use for dimension bigger than 9, because it uses the basis() function

import numpy as np
import numpy
import sys
import os
import scipy.linalg
from six import print_

# add the parent directory, so we find our iisignature build if it was built --inplace
sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
import iisignature

# the following 3 functions for interpreting basis strings
# are copied from the tests


# text, index -> (either number or [res, res]), newIndex
def parseBracketedExpression(text, index):
    if text[index] == "[":
        left, m = parseBracketedExpression(text, index + 1)
        right, n = parseBracketedExpression(text, m + 1)
        return [left, right], n + 1
    else:
        n = 0
        while (
            n < len(text)
            and text[index + n] in ["1", "2", "3", "4", "5", "6", "7", "8", "9"]
        ):  # this should always just happen once if input is a bracketed expression of letters
            n = n + 1
        return int(text[index : index + n]), index + n


# print (parseBracketedExpression("[23,[2,[[22,1],2]]]",0))


# bracketed expression, dim -> numpy array of its value, depth
def multiplyOut(expn, dim):
    if isinstance(expn, list):
        left, leftDepth = multiplyOut(expn[0], dim)
        right, rightDepth = multiplyOut(expn[1], dim)
        a = numpy.outer(left, right).flatten()
        b = numpy.outer(right, left).flatten()
        return a - b, leftDepth + rightDepth
    else:
        a = numpy.zeros(dim)
        a[expn - 1] = 1
        return a, 1


# string of bracketed expression, dim -> numpy array of its value, depth
# for example:
# valueOfBracket("[1,2]",2) is ([0,1,-1,0],2)
def valueOfBracket(text, dim):
    return multiplyOut(parseBracketedExpression(text, 0)[0], dim)


def getMatrices(s):
    info = iisignature.info(s)
    m = info["level"]
    d = info["dimension"]
    values = [[] for i in range(m)]
    for exp in iisignature.basis(s):
        vec, level = valueOfBracket(exp, d)
        values[level - 1].append(vec)
    out = []
    for v in values:

        def f(x, y):
            return numpy.inner(v[x], v[y])

        out.append(numpy.fromfunction(numpy.vectorize(f), (len(v), len(v)), dtype=int))
    #        store=[]
    #        for x1 in v:
    #            store.append([])
    #            for x2 in v:
    #                store[-1].append(numpy.inner(x1,x2))
    #        out.append(numpy.array(store))
    return out


def demo():
    # 1. generate some dummy data
    d = 3
    m = 6
    path1 = np.random.uniform(size=(20, d))
    path2 = np.random.uniform(size=(20, d))
    paths = (path1, path2)
    s = iisignature.prepare(d, m)

    # 2. print the dot product of the log signatures in tensor space
    # (The "x" means that the log signature is returned in tensor space.)
    expandedLogSig1, expandedLogSig2 = (iisignature.logsig(i, s, "x") for i in paths)
    target = np.inner(expandedLogSig1, expandedLogSig2)
    print_("Target:", float(target))

    # 3. use getMatrices to act on the log signatures expressed in a basis
    #    and get the same answer.
    matrices = getMatrices(s)
    logsig1, logsig2 = (iisignature.logsig(i, s) for i in paths)
    adjustment = scipy.linalg.block_diag(*matrices)
    # print(np.dot(adjustment,logsig2).shape)
    print_("We get:", float(np.inner(logsig1, np.dot(adjustment, logsig2))))


if __name__ == "__main__":
    demo()



# This module defines a theano op called Sig,
# which just does iisignature.sig,
# one called LogSig,
# which just does iisignature.logsig,
# one called SigJoin,
# which just does iisignature.sigjoin,
# and one called SigScale,
# which just does iisignature.sigscale.

import theano
import numpy as np
import iisignature

# Todo: you can use x.ndim in make_node to change behaviour based on shapes of the input.
# could make these functions optionally batched

# Turn on to throw an exception on nan
nancheck = False


def contains_nan(x):
    return np.isnan(x).any()


# Turn on to throw exceptions in some places when called on a non-contiguous array
# At the moment, iisignature functions will do extra copying when this happens.
# If it happens a lot, we could adapt the code.
report_contig = False
if report_contig:

    def iscontig(arr):
        return arr.flags["C_CONTIGUOUS"]

    def contig_check(a):
        for i, aa in enumerate(a):
            if not iscontig(aa):
                raise RuntimeError(str(i + 1) + "th argument is not contiguous")


# This is a theano Op which wraps iisignature.siglength .
# It is used to implement shape inference of Sig.
class SigLength_op(theano.Op):
    __props__ = ()

    def make_node(self, d, m):
        d = theano.tensor.as_tensor_variable(d)
        m = theano.tensor.as_tensor_variable(m)
        return theano.Apply(self, [d, m], [theano.tensor.iscalar()])

    def perform(self, node, inp, out):
        # do I really need to create an array here?
        out[0][0] = np.array(iisignature.siglength(inp[0], inp[1]), dtype="int32")

    def infer_shape(self, node, shapes):
        return [[]]


SigLength = SigLength_op()


# This is a theano Op which wraps iisignature.logsiglength .
# It could be used to implement shape inference of LogSig, but isn't.
class LogSigLength_op(theano.Op):
    __props__ = ()

    def make_node(self, d, m):
        d = theano.tensor.as_tensor_variable(d)
        m = theano.tensor.as_tensor_variable(m)
        return theano.Apply(self, [d, m], [theano.tensor.iscalar()])

    def perform(self, node, inp, out):
        out[0][0] = np.array(iisignature.logsiglength(inp[0], inp[1]), dtype="int32")

    def infer_shape(self, node, shapes):
        return [[]]


LogSigLength = LogSigLength_op()


# This is a theano Op which wraps iisignature.sigbackprop .
# It is used in the grad method of Sig.
# Ideally we would have an optimization to sum a variable with this in place
# - like backward in torch
class SigGrad_op(theano.Op):
    __props__ = ()

    def infer_shape(self, node, shapes):
        return [shapes[1]]

    def make_node(self, s, x, m):
        s = theano.tensor.as_tensor_variable(s)
        x = theano.tensor.as_tensor_variable(x)
        m = theano.tensor.as_tensor_variable(m)
        outT = theano.tensor.TensorType("float32", [False] * x.ndim)()
        return theano.Apply(self, inputs=[s, x, m], outputs=[outT])

    def perform(self, node, inputs_storage, out):
        s = inputs_storage[0]
        x = inputs_storage[1]
        m = inputs_storage[2]
        out[0][0] = iisignature.sigbackprop(s, x, m)


SigGrad = SigGrad_op()


# This is a theano Op which wraps iisignature.sig .
class Sig_op(theano.Op):
    __props__ = ()

    def infer_shape(self, node, shapes):
        return [shapes[0][:-2] + (SigLength(shapes[0][-1], node.inputs[1]),)]

    def make_node(self, x, m):
        x_ = theano.tensor.as_tensor_variable(x)
        m = theano.tensor.as_tensor_variable(m)
        outT = theano.tensor.TensorType("float32", [False] * (x.ndim - 1))()
        return theano.Apply(self, inputs=[x_, m], outputs=[outT])

    def perform(self, node, inputs_storage, outputs_storage):
        x = inputs_storage[0]
        m = inputs_storage[1]
        outputs_storage[0][0] = iisignature.sig(x, m)

    def grad(self, inputs, g):
        return [
            SigGrad(g[0], inputs[0], inputs[1]),
            theano.gradient.grad_undefined(self, 1, inputs[1]),
        ]


Sig = Sig_op()

# There is no smooth way to pass information which is known in one instance of the
# Op, like the prepared object in LogSig, to the Op's grad() method. node is
# not one of the parameters of grad. Thus we store the prepared objects in
# this global variable, and have its index as a constant input to the apply node.
_prepared_obeject_store = []


# This is a theano Op which wraps iisignature.logsigbackprop .
# The arguments are a bit weird: if you want to use this yourself
# you need to append the prepared object and method as a tuple to
# _prepared_obeject_store and pass the index as an input.
class LogSigGrad_op(theano.Op):
    __props__ = ()

    def infer_shape(self, node, shapes):
        return [shapes[1]]

    def make_node(self, grad_output, x, index):
        grad_output = theano.tensor.as_tensor_variable(grad_output)
        x = theano.tensor.as_tensor_variable(x)
        outT = theano.tensor.TensorType("float32", [False] * x.ndim)()
        return theano.Apply(self, inputs=[grad_output, x, index], outputs=[outT])

    def perform(self, node, inputs_storage, out):
        grad_output = inputs_storage[0]
        x = inputs_storage[1]
        s, method = _prepared_obeject_store[inputs_storage[2]]
        out[0][0] = iisignature.logsigbackprop(grad_output, x, s, method)


LogSigGrad = LogSigGrad_op()


# This is a theano Op which wraps iisignature.logsig
class LogSig_op(theano.Op):
    __props__ = ()

    def infer_shape(self, node, shapes):
        # d=shapes[0][-1]
        s, method = _prepared_obeject_store[node.inputs[1].get_value()]
        m = iisignature.info(s)["level"]
        d = iisignature.info(s)["dimension"]
        if "x" in method or "X" in method:
            length = iisignature.siglength(d, m)
        else:
            length = iisignature.logsiglength(d, m)
        return [shapes[0][:-2] + (length,)]

    def make_node(self, x, s, method=""):
        x_ = theano.tensor.as_tensor_variable(x)
        outT = theano.tensor.TensorType("float64", [False] * (x.ndim - 1))()
        index = theano.shared(len(_prepared_obeject_store))
        _prepared_obeject_store.append((s, method))
        return theano.Apply(self, inputs=[x_, index], outputs=[outT])

    def perform(self, node, inputs_storage, outputs_storage):
        x = inputs_storage[0]
        s, method = _prepared_obeject_store[inputs_storage[1]]
        outputs_storage[0][0] = iisignature.logsig(x, s, method)

    def grad(self, inputs, g):
        return [
            LogSigGrad(g[0], inputs[0], inputs[1]),
            theano.gradient.grad_undefined(self, 1, inputs[1]),
        ]


LogSig = LogSig_op()


# This is a theano Op which wraps iisignature.sigjoinbackprop .
# It has multiple outputs.
class SigJoinGrad_op(theano.Op):
    __props__ = ()

    def infer_shape(self, node, shapes):
        return [shapes[1], shapes[2], None]

    def make_node(self, s, x, y, m, fixed):
        s = theano.tensor.as_tensor_variable(s)
        x = theano.tensor.as_tensor_variable(x)
        y = theano.tensor.as_tensor_variable(y)
        m = theano.tensor.as_tensor_variable(m)
        fixed = theano.tensor.as_tensor_variable(fixed)
        outT = theano.tensor.TensorType("float32", [False] * (x.ndim))
        return theano.Apply(
            self,
            inputs=[s, x, y, m, fixed],
            outputs=[outT(), outT(), theano.tensor.dscalar()],
        )

    def perform(self, node, inputs_storage, out):
        s = inputs_storage[0]
        x = inputs_storage[1]
        y = inputs_storage[2]
        m = inputs_storage[3]
        fixed = inputs_storage[4]
        if report_contig:
            contig_check([s, x, y])
        o = iisignature.sigjoinbackprop(s, x, y, m, fixed)
        if nancheck:
            if contains_nan(s):
                raise RuntimeError("nan in s")
            if contains_nan(x):
                raise RuntimeError("nan in x")
            if contains_nan(y):
                raise RuntimeError("nan in y")
            if contains_nan(o[0]) or contains_nan(o[1]):
                raise RuntimeError("nan in output")
        out[0][0] = o[0]
        out[1][0] = o[1]
        out[2][0] = np.array(0.0 if np.isnan(fixed) else o[2], dtype="float64")


SigJoinGrad = SigJoinGrad_op()


# This is a theano Op which wraps iisignature.sigjoin .
class SigJoin_op(theano.Op):
    __props__ = ()

    def infer_shape(self, node, shapes):
        return [shapes[0]]

    def make_node(self, x, y, m, fixed=float("nan")):
        x = theano.tensor.as_tensor_variable(x)
        y = theano.tensor.as_tensor_variable(y)
        m = theano.tensor.as_tensor_variable(m)
        fixed = theano.tensor.as_tensor_variable(fixed)
        outT = theano.tensor.TensorType("float32", [False] * (x.ndim))
        return theano.Apply(self, inputs=[x, y, m, fixed], outputs=[outT()])

    def perform(self, node, inputs_storage, outputs_storage):
        x = inputs_storage[0]
        y = inputs_storage[1]
        m = inputs_storage[2]
        fixed = inputs_storage[3]
        outputs_storage[0][0] = iisignature.sigjoin(x, y, m, fixed)
        if nancheck:
            if contains_nan(x):
                raise RuntimeError("nan in x")
            if contains_nan(y):
                raise RuntimeError("nan in y")
            if contains_nan(outputs_storage[0][0]):
                raise RuntimeError("nan in output")

    def grad(self, inputs, g):
        gg = SigJoinGrad(g[0], inputs[0], inputs[1], inputs[2], inputs[3])
        return [gg[0], gg[1], theano.gradient.grad_undefined(self, 2, inputs[2]), gg[2]]


SigJoin = SigJoin_op()


# This is a theano Op which wraps iisignature.sigscalebackprop .
# It has two outputs.
class SigScaleGrad_op(theano.Op):
    __props__ = ()

    def infer_shape(self, node, shapes):
        return [shapes[1], shapes[2]]

    def make_node(self, s, x, y, m):
        s = theano.tensor.as_tensor_variable(s)
        x = theano.tensor.as_tensor_variable(x)
        y = theano.tensor.as_tensor_variable(y)
        m = theano.tensor.as_tensor_variable(m)
        outT = theano.tensor.TensorType("float32", [False] * (x.ndim))
        return theano.Apply(self, inputs=[s, x, y, m], outputs=[outT(), outT()])

    def perform(self, node, inputs_storage, out):
        s = inputs_storage[0]
        x = inputs_storage[1]
        y = inputs_storage[2]
        m = inputs_storage[3]
        if report_contig:
            contig_check([s, x, y])
        o = iisignature.sigscalebackprop(s, x, y, m)
        if nancheck:
            if contains_nan(s):
                raise RuntimeError("nan in s")
            if contains_nan(x):
                raise RuntimeError("nan in x")
            if contains_nan(y):
                raise RuntimeError("nan in y")
            if contains_nan(o[0]) or contains_nan(o[1]):
                raise RuntimeError("nan in output")
        out[0][0] = o[0]
        out[1][0] = o[1]


SigScaleGrad = SigScaleGrad_op()


# This is a theano Op which wraps iisignature.sigscale .
class SigScale_op(theano.Op):
    __props__ = ()

    def infer_shape(self, node, shapes):
        return [shapes[0]]

    def make_node(self, x, y, m):
        x = theano.tensor.as_tensor_variable(x)
        y = theano.tensor.as_tensor_variable(y)
        m = theano.tensor.as_tensor_variable(m)
        outT = theano.tensor.TensorType("float32", [False] * (x.ndim))
        return theano.Apply(self, inputs=[x, y, m], outputs=[outT()])

    def perform(self, node, inputs_storage, outputs_storage):
        x = inputs_storage[0]
        y = inputs_storage[1]
        m = inputs_storage[2]
        outputs_storage[0][0] = iisignature.sigscale(x, y, m)
        if nancheck:
            if contains_nan(x):
                raise RuntimeError("nan in x")
            if contains_nan(y):
                raise RuntimeError("nan in y")
            if contains_nan(outputs_storage[0][0]):
                raise RuntimeError("nan in output")

    def grad(self, inputs, g):
        gg = SigScaleGrad(g[0], inputs[0], inputs[1], inputs[2])
        return [gg[0], gg[1], theano.gradient.grad_undefined(self, 2, inputs[2])]


SigScale = SigScale_op()


# http://deeplearning.net/software/theano/extending/extending_theano_gpu.html
# http://theano.readthedocs.io/en/rel-0.6rc3/tutorial/extending_theano.html



import os
import sys
import numpy as np

# add the parent directory, so we find our iisignature build if it was built --inplace
sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
import iisignature_match_esig as tosig

d = 3
m = 4
path = np.random.uniform(size=(10, d))

print(tosig.stream2sig(path, m))
print(tosig.stream2logsig(path, m))
print(tosig.sigkeys(d, m))
print(tosig.logsigkeys(d, m))
print(tosig.sigdim(d, m))
print(tosig.logsigdim(d, m))



# This shows paths which are made of traversing a figure of eight,
# you seem to be able to get arbitrarily (Fibonaccily?) many levels of the signature to be 0.

import sys
import os
import numpy as np
from math import pi

# add the parent directory, so we find our iisignature build if it was built --inplace
sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
import iisignature

np.set_printoptions(suppress=True, precision=6)


fractions = np.arange(1, 101) / 100.0

# This is a unit circle centred on [0.5,0]
positivePetal = np.column_stack(
    [1 - np.cos(fractions * pi * 2), np.sin(fractions * pi * 2)]
)

petals = [i * positivePetal for i in [[1, 1], [1, -1], [-1, 1], [-1, -1]]]
#   2<-   #                     #    ->0   #
#               #              #                  #

#                     #        #                        #

#                           #   #                             #
# #

#                             O

#                            # #                              #
#   #

#                        #       #                         #

#               #              #                   #
#   3<- #                       #    ->1    #


# USER: leave uncommented the version of seq which you care about.
# circle - nonzero level 2
seq = [1]

# figure of eight - nonzero level 3
seq = [1, 3]
seq = [2, 0]

# nonzero level 5
seq = [1, 3, 0, 2]
# seq=[3,1,2,0]
# seq=[2,0,3,1]
# seq=[0,2,1,3]

# nonzero level 8
seq = [1, 3, 0, 2, 2, 0, 3, 1]
# seq=[3,1,2,0,0,2,1,3]
# seq=[2,0,3,1,1,3,0,2]
# seq=[0,2,1,3,3,1,2,0]

# nonzero level 13
seq = [1, 3, 0, 2, 2, 0, 3, 1, 3, 1, 2, 0, 0, 2, 1, 3]
# seq=[3,1,2,0,0,2,1,3,1,3,0,2,2,0,3,1]
# seq=[0,2,1,3,3,1,2,0,2,0,3,1,1,3,0,2]
# seq=[2,0,3,1,1,3,0,2,0,2,1,3,3,1,2,0]

# zero up to
# seq=[1,3,0,2,2,0,3,1,3,1,2,0,0,2,1,3,0,2,1,3,3,1,2,0,2,0,3,1,1,3,0,2]

# The pattern from one to the next seems to be:
#  0 -> 2 0
#  1 -> 1 3
#  2 -> 3 1
#  3 -> 0 2

total = [[(0, 0)]]
for i in range(1, len(seq)):
    s = {seq[i - 1], seq[i]}
    if s in [{0, 1}, {2, 3}]:
        raise RuntimeError("treelike")
for i in seq:
    offset = np.array([0, 0])  # change this to [2,0] to display petals separately
    startpoint = total[-1][-1] + offset
    if i in [0, 1, 2, 3]:
        total.append(startpoint + petals[i])
    else:
        raise "whoops"

path = np.vstack(total)
print("Maximum absolute value of sig elements in each level:")
for i, j in enumerate(iisignature.sig(path, 14, 1)):
    print("level {:2}:".format(i + 1), np.max(np.abs(j)))

if 0:
    import matplotlib.pyplot as plt
    import matplotlib as mpl

    # colors = np.arange(path.
    # plt.plot(path[:,0],path[:,1])
    plt.scatter(
        path[:, 0],
        path[:, 1],
        c=np.arange(path.shape[0]),
        norm=mpl.colors.Normalize(vmin=0, vmax=path.shape[0]),
    )
    plt.show()


# A sort of Thue Morse chain of fig eights



# ruff: noqa
# This allows you to draw paths freehand with the mouse on the canvas, and
# see their log signatures printed on the console.
# based on http://www.tkdocs.com/tutorial/canvas.html

from tkinter import *
import sys
import os
import numpy as np

# add the parent directory, so we find our iisignature build if it was built --inplace
sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
import iisignature

np.set_printoptions(suppress=True, precision=6)

m = 3
s = iisignature.prepare(2, m)

lastx, lasty = 0, 0
points = []

root = Tk()
root.columnconfigure(0, weight=1)
root.rowconfigure(0, weight=1)

canvas = Canvas(root)


def xy(event):
    global lastx, lasty, points
    lastx, lasty = event.x, event.y
    points = [[lastx, -lasty]]


def addLine(event):
    global lastx, lasty
    canvas.create_line((lastx, lasty, event.x, event.y))
    lastx, lasty = event.x, event.y
    points.append([lastx, -lasty])


def doneStroke(event):
    # scaled_points = points / np.array([canvas.winfo_width(), canvas.winfo_height()])
    scaled_points = points / np.array(max(canvas.winfo_width(), canvas.winfo_height()))
    # print(iisignature.sig(points,m))
    # print(iisignature.logsig(points,s))
    # print(iisignature.sig(scaled_points,m))
    print(iisignature.logsig(scaled_points, s))


canvas.grid(column=0, row=0, sticky=(N, W, E, S))
canvas.bind("<Button-1>", xy)
canvas.bind("<B1-Motion>", addLine)
canvas.bind("<B1-ButtonRelease>", doneStroke)
root.mainloop()



import torch
from torch.autograd import Variable
import torch.nn as nn
# from torch.nn.modules.module import Module

from iisignature_torch import SigJoin
import iisignature
import math

# This is a pytorch example of a recurrent layer which uses Signature internally.
# It is not identical to the example in iisignature_recurrent_keras, e.g. in that there are
# more bias parameters, no activation, and less control over the initialisation.
# No assertion is made that this model is a good idea, or that this code is idiomatic pytorch.


class RecurrentSig(nn.Module):
    def __init__(
        self,
        input_size,
        n_units,
        sig_level=2,
        train_time_lapse=True,
        initial_time_lapse=0.1,
        output_signatures=False,
        use_signatures=True,
    ):
        super(RecurrentSig, self).__init__()
        self.sig_level = sig_level
        self.sigsize = iisignature.siglength(2, sig_level)
        self.n_units = n_units
        self.units = n_units * (self.sigsize + 1) if output_signatures else n_units
        self.train_time_lapse = train_time_lapse
        # self.initial_time_lapse = initial_time_lapse
        self.output_signatures = output_signatures
        self.use_signatures = use_signatures

        saved_data_length = self.n_units * (
            1 + self.sigsize if self.use_signatures else 1
        )
        self.W = nn.Linear(input_size, self.n_units)
        self.U = nn.Linear(saved_data_length, self.n_units)
        if self.train_time_lapse:
            self.log_timelapse = nn.parameter.Parameter(
                torch.FloatTensor([math.log(initial_time_lapse)])
            )
        else:
            self.time_lapse = Variable(torch.FloatTensor([initial_time_lapse]))

    def initHidden(self, batch_size):
        lengths = (self.n_units * self.sigsize, self.n_units)
        a = Variable(torch.zeros(batch_size, lengths[0]))
        b = Variable(torch.zeros(batch_size, lengths[1]))
        return a, b

    def forward(self, input, states):
        prev_sigs = states[0]  # (batch,n_units*sigsize)
        prev_sigs_ = prev_sigs.view(-1, self.sigsize)  # (batch*n_units,sigsize)
        prev_states = states[1]
        h = self.W(input)
        prev_states_as_memory = (
            torch.cat([prev_sigs, prev_states], 1)
            if self.use_signatures
            else prev_states
        )

        raw_output = h + self.U(prev_states_as_memory)

        displacements = (raw_output - prev_states).view(-1, 1)
        time_lapse = (
            torch.exp(self.log_timelapse) if self.train_time_lapse else self.time_lapse
        )
        sigs = SigJoin(prev_sigs_, displacements, self.sig_level, time_lapse)
        sigs = sigs.view(-1, self.n_units * self.sigsize)
        if self.output_signatures:
            output = torch.cat([sigs, raw_output], 1)
        else:
            output = raw_output
        return output, [sigs, raw_output]


if __name__ == "__main__":
    # Just a sanity check
    r = RecurrentSig(1, 1)
    hidden = r.initHidden(3)
    inp = Variable(torch.randn(3, 1))
    result, hidden = r(inp, hidden)
    print(result)



# In this simple example, we demonstrate using the signature functionality from iisignature
# by attempting to learn the points of a path
# from its signature via gradient descent.
# The starting point is another path whose first point matches the target

import os

# os.environ["THEANO_FLAGS"]="floatX=float32,device=cpu,optimizer=fast_compile"
# os.environ["THEANO_FLAGS"]="floatX=float32,device=cpu,mode=DebugMode"
os.environ["THEANO_FLAGS"] = "floatX=float32,device=cpu"
import theano
import numpy
import sys
import six.moves

# add the parent directory, so we find our iisignature build if it was built --inplace
sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
import iisignature

from iisignature_theano import LogSig, Sig


# 1: SETUP VARIABLES
dim = 2
level = 3
pathlength = 4
timedim = False
useLogSig = True
s = iisignature.prepare(dim, level)

numpy.random.seed(51)
target = numpy.random.uniform(size=(pathlength, dim)).astype("float32")
# target = numpy.cumsum(2*(target-0.5),0)#makes it more random-walk-ish, less like a scribble

targetSig = (
    iisignature.logsig(target, s) if useLogSig else iisignature.sig(target, level)
)
start = numpy.random.uniform(size=(pathlength, dim)).astype("float32")
start[0, :] = target[0, :]
learnable_mask = numpy.ones((pathlength, dim)).astype("float32")
learnable_mask[0, :] = 0  # to stop the starting point being modified
if timedim:
    for i in six.moves.xrange(pathlength):
        target[i, 0] = start[i, 0] = i * 0.2
        learnable_mask[i, 0] = 0
learning_rate_decay = 1.003
initial_learning_rate = 1.2
momentum = 0.95

# 2: DEFINE THEANO STUFF

learning_rate = theano.shared(numpy.float32(initial_learning_rate), "learning_rate")
path = theano.shared(start, "path")
grad_avg = theano.shared(numpy.zeros_like(start, dtype="float32"), "grad_avg")
sig = LogSig(path, s) if useLogSig else Sig(path, level)
cost = theano.tensor.mean(theano.tensor.sqr(sig - targetSig))
new_grad_avg = (momentum * grad_avg) + (1 - momentum) * theano.grad(cost, path)
use_old = False
grad_to_use = grad_avg if use_old else new_grad_avg
ff = theano.function(
    [],
    [cost],
    updates=[
        (path, path - learning_rate * learnable_mask * grad_to_use),
        (grad_avg, new_grad_avg),
        (learning_rate, learning_rate * learning_rate_decay),
    ],
)

# 3: GO

numpy.set_printoptions(suppress=True)
print("target:")
print(target)
print("start:")
print(start)
for i in six.moves.xrange(581):
    stepcost = ff()  # one step of gradient descent
    if i % 18 == 0:
        print("step " + str(i) + " cost: " + str(ff()[0].item()))
print("end:")
print(path.get_value())
# a=iisignature.sig(start_.get_value(),level)
# print (numpy.vstack([a,targetSig,a-targetSig]).transpose())

##http://deeplearning.net/software/theano/extending/extending_theano_gpu.html
##http://theano.readthedocs.io/en/rel-0.6rc3/tutorial/extending_theano.html



import numpy


# takes a list of nonnegative integers to a path representation
# as an axis path
# e.g. [0,1,0] -> [[0,0],[1,0],[1,1],[2,1]]
def ex(a):
    n = max(a) + 1
    out = [[0.0] * n]
    for i in a:
        v = out[-1][:]
        v[i] = v[i] + 1
        out.append(v)
    return numpy.array(out)



# In this simple example, we demonstrate using the signature functionality from iisignature
# by attempting to learn the points of a path
# from its signature via gradient descent.
# The starting point is another path whose first point matches the target

import os

# os.environ["THEANO_FLAGS"]="floatX=float32,device=cpu,optimizer=fast_compile"
# os.environ["THEANO_FLAGS"]="floatX=float32,device=cpu,mode=DebugMode"
os.environ["THEANO_FLAGS"] = "floatX=float32,device=cpu"
import theano
import numpy
import sys
import theano.tensor as T

# add the parent directory, so we find our iisignature build if it was built --inplace
sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
import iisignature

from iisignature_theano import LogSig


# 1: SETUP VARIABLES
dim = 2
level = 3
pathlength = 4
timedim = False
useLogSig = True
s = iisignature.prepare(dim, level)

numpy.random.seed(51)
target = numpy.random.uniform(size=(pathlength, dim)).astype("float32")

a = T.dmatrix("a")
b = T.dmatrix("b")

# c=LogSigGrad(b,a,s)
# f=theano.function([a,b],c)
# print (f(target,target))

c = LogSig(b, s)
grad = theano.grad(theano.tensor.sum(c), b)
f = theano.function([b], [c, grad])
o = f(target)

print(o[0])
print(o[1])



from setuptools import setup, Extension
import os
import sys

version = "0.24"
headers = [
    "bch",
    "calcSignature",
    "logsig",
    "logSigLength",
    "makeCompiledFunction",
    "rotationalInvariants",
    "readBCHCoeffs",
]

args = []
link_args = []
if os.name == "posix":
    args = ["-std=c++14"]

if sys.platform == "darwin" and "MACOSX_DEPLOYMENT_TARGET" not in os.environ:
    if "IISIGNATURE_MACOSX_DONOTBECLEVER" not in os.environ:
        args.append("-mmacosx-version-min=10.9")
        link_args.append("-mmacosx-version-min=10.9")
        # os.environ["MACOSX_DEPLOYMENT_TARGET"]="10.9"


# https://github.com/pybind/python_example/blob/2ed5a68759cd6ff5d2e5992a91f08616ef457b5c/setup.py#L9
# and
# https://stackoverflow.com/questions/54117786/add-numpy-get-include-argument-to-setuptools-without-preinstalled-numpy
class get_numpy_include(object):
    def __str__(self):
        import numpy

        return numpy.get_include()


xtn = Extension(
    "iisignature",
    ["src/pythonsigs.cpp"],
    extra_compile_args=args,
    extra_link_args=link_args,
    define_macros=[("VERSION", version)],
    include_dirs=[get_numpy_include()],
    depends=["src/" + i + ".hpp" for i in headers],
)


def readme():
    with open("README.rst") as f:
        return f.read()


setup(
    name="iisignature",
    version=version,
    ext_modules=[xtn],
    description="Iterated integral signature calculations",
    long_description=readme(),
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Financial and Insurance Industry",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Mathematics",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here.
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.8",
    url="https://github.com/bottler/iisignature",
    author="Jeremy Reizenstein",
    author_email="j.f.reizenstein@warwick.ac.uk",
    keywords=["signature", "logsignature"],
    license="MIT",
    packages=["iisignature_data"],
    test_suite="tests.my_module_suite",
    # data_files=[("iisignature_data",["src/bchLyndon20.dat"])],
    package_data={"iisignature_data": ["bchLyndon20.dat"]},
    zip_safe=False,
    install_requires=[
        "numpy>=1.17"
    ],  # For now, this is redundant as we've died above here on an import statement
    setup_requires=["numpy>=1.17"],
)
# should add source url from github like at https://www.python.org/dev/peps/pep-0426/



# This module defines a PyTorch function called Sig,
#  which just does iisignature.sig,
# one called LogSig,
#  which just does iisignature.logsig,
# one called SigJoin,
#  which just does iisignature.sigjoin,
# and one called SigScale,
#  which just does iisignature.sigscale.
import torch
from torch.autograd import Function

import iisignature


class SigFn(Function):
    @staticmethod
    def forward(ctx, X, m):
        result = iisignature.sig(X.detach().numpy(), m)
        ctx.save_for_backward(X)
        ctx.m = m
        return torch.FloatTensor(result)

    @staticmethod
    def backward(ctx, grad_output):
        (X,) = ctx.saved_tensors
        m = ctx.m
        result = iisignature.sigbackprop(grad_output.numpy(), X.detach().numpy(), m)
        return torch.FloatTensor(result), None


class LogSigFn(Function):
    @staticmethod
    def forward(ctx, X, s, method):
        result = iisignature.logsig(X.detach().numpy(), s, method)
        ctx.save_for_backward(X)
        ctx.s = s
        ctx.method = method
        return torch.FloatTensor(result)

    @staticmethod
    def backward(ctx, grad_output):
        (X,) = ctx.saved_tensors
        s = ctx.s
        method = ctx.method
        g = grad_output.numpy()
        result = iisignature.logsigbackprop(g, X.detach().numpy(), s, method)
        return torch.FloatTensor(result), None, None


class SigJoinFn(Function):
    @staticmethod
    def forward(ctx, X, Y, m):
        result = iisignature.sigjoin(X.detach().numpy(), Y.detach().numpy(), m)
        ctx.save_for_backward(X, Y)
        ctx.m = m
        return torch.FloatTensor(result)

    @staticmethod
    def backward(ctx, grad_output):
        X, Y = ctx.saved_tensors
        m = ctx.m
        result = iisignature.sigjoinbackprop(
            grad_output.numpy(), X.detach().numpy(), Y.detach().numpy(), m
        )
        return torch.FloatTensor(result[0]), torch.FloatTensor(result[1]), None


class SigJoinFixedFn(Function):
    @staticmethod
    def forward(ctx, X, Y, m, fixed):
        result = iisignature.sigjoin(
            X.detach().numpy(), Y.detach().numpy(), m, fixed.detach().numpy()
        )
        ctx.save_for_backward(X, Y, fixed)
        ctx.m = m
        return torch.FloatTensor(result)

    @staticmethod
    def backward(ctx, grad_output):
        X, Y, fixed = ctx.saved_tensors
        m = ctx.m
        result = iisignature.sigjoinbackprop(
            grad_output.numpy(),
            X.detach().numpy(),
            Y.detach().numpy(),
            m,
            fixed.detach().numpy(),
        )
        return (
            torch.FloatTensor(result[0]),
            torch.FloatTensor(result[1]),
            None,
            torch.FloatTensor([result[2]]),
        )


class SigScaleFn(Function):
    @staticmethod
    def forward(ctx, X, Y, m):
        result = iisignature.sigscale(X.detach().numpy(), Y.detach().numpy(), m)
        ctx.save_for_backward(X, Y)
        ctx.m = m
        return torch.FloatTensor(result)

    @staticmethod
    def backward(ctx, grad_output):
        X, Y = ctx.saved_tensors
        m = ctx.m
        result = iisignature.sigscalebackprop(
            grad_output.numpy(), X.detach().numpy(), Y.detach().numpy(), m
        )
        return torch.FloatTensor(result[0]), torch.FloatTensor(result[1]), None


def Sig(X, m):
    return SigFn.apply(X, m)


def LogSig(X, s, method=""):
    return LogSigFn.apply(X, s, method)


def SigJoin(X, y, m, fixed=None):
    if fixed is not None:
        return SigJoinFixedFn.apply(X, y, m, fixed)
    return SigJoinFn.apply(X, y, m)


def SigScale(X, y, m):
    return SigScaleFn.apply(X, y, m)
